####            CLASSES FOR B KING                    ####################################################
##   Those are the classes used with the game.          ##################################################
##   Girl class and functions are in separate files.    ##################################################
##                                                      ##################################################


init -2 python:
    #<Chris12 PackState>
    import hashlib
    import datetime
    #</Chris12 PackState>

## PRIMARY CLASSES ##

    class Game(object):
        
        """This class keeps track of various story related variables and methods."""
        
        def __init__(self):
            
            self.started = False
            self.chapter = 1
            self.token = 0
            self.girl_id_generated = 0
            self.free_girls = []
            self.kidnapped = []
            self.goals = (Goal("gold", 1000),)
            self.set_task(self.get_goal_description())
            self.seen_goal_message = False
            self.blocked_districts = []
            self.active_mods = {}
            self.track_dict = defaultdict(int)
            self.last_pic = {"tags": [], "and_tags": [], "not_tags": [], "attempts": 0}
            self.load_pics()
            self.effects = []
            self.effect_dict = defaultdict(list)
            self.customer_preference_weight = defaultdict(int)
#            self.set_difficulty("normal")
            self.__filesdict_timestamp = datetime.datetime.now() #<Chris12 AutoRepair />
        
        # Difficulty
        
        def set_difficulty(self, diff, custom_diff=None): # Where diff is a string appearing in the 'diff_dict' dictionary
            
            # custom_diff must be a dictionary including all difficulty settings
            
            if diff == "custom":
                self.diff = custom_diff
                
            else:
                self.diff = diff_dict[diff]
        
        # Effects (game effects apply in the city and affect city girls)
        
        def get_effect(self, type, target): 
            return get_effect(self, type, target)
                        
        def add_effects(self, effects, apply_boost=False, spillover=False, expires = False):
            return add_effects(self, effects, apply_boost=apply_boost, spillover=spillover, expires=expires)
            
        def remove_effects(self, effects):
            remove_effects(self, effects) 
        
        def load_pics(self): # Those are the default pics for the game
            
            # Resetting pictures
            
            self.pics = []
            
            # Creating pictures
            
            for file in [f for f in renpy.list_files() if f.startswith("default/") and is_imgfile(f)]:
                
                file_name = file.split("/")[-1]
                
                pic = Picture(file_name, file)
                
                self.pics.append(pic)
                
                # Tracing untagged pics for debugging
                
                if pic.tags == []:
                    untagged_pics.append(pic.path)
        
        def track(self, key, value):
            self.track_dict[key] += value
            
        def check(self, key):
            return self.track_dict[key]
        
        def start_mods(self):
            # Init active mods (mods are stored within the game object, unless they get overwritten after an update)
            for mod in detected_mods.values():
                if mod.active:
                    self.activate_mod(mod)
        
        def update_mods(self):
            
            # Checks if a saved game's mods have been deleted or changed
            
            for mod in self.active_mods.values():
                
#                renpy.say("", str(mod))
                
                if mod.name not in detected_mods.keys():
                    if renpy.call_screen("yes_no", mod.full_name + " couldn't be found. Would you like to deactivate this mod for this game (recommended)?"):
                        self.deactivate_mod(mod)
                
                elif mod.check_for_updates():
                    if renpy.call_screen("yes_no", "A different version of this mod: " + mod.name + " has been found (" + str(mod.version) + "). Would you like to reset this mod for this game (recommended)?"):
                        self.deactivate_mod(mod)
                        self.activate_mod(detected_mods[mod.name])
                
                elif not persistent.mods[mod.name]["active"]:
                    if renpy.call_screen("yes_no", mod.full_name + " has been deactivated. Would you like to deactivate this mod for this game?"):
                        self.deactivate_mod(mod)
            
            # Checks if a new mod has been activated
            
            for name, mod in detected_mods.items():
                if mod.active:
                    if name not in self.active_mods.keys():
                        if renpy.call_screen("yes_no", "A new mod has been activated: " + mod.full_name + ". Would you like to activate this mod for this game?"):
                            self.activate_mod(mod)
                                    
            updated_games[self] = True # To do: Check if it works or needs a function
            
        def activate_mod(self, mod):
            self.active_mods[mod.name] = mod

            if mod.night_label:
                daily_events.append(StoryEvent(label=mod.night_label, type="night", once=False))
                
            renpy.notify("\n" + mod.name + " has been activated.")
            
#            renpy.say("", str(self.active_mods[mod.name]))
            
            if mod.init_label:
#                try:
                renpy.call(mod.init_label)
#                except:
#                    renpy.say("System", event_color["bad"] % ("Failure to start " + mod.name) + " (calling " + mod.init_label + " label failed).")
            
            
        def deactivate_mod(self, mod):
            try:
                del self.active_mods[mod.name]
            except:
                renpy.say("System", event_color["bad"] % ("Failure to deactivate " + mod.name))
            
            if mod.night_label:
                for ev in daily_events:
                    if ev.name == mod.night_label:
                        daily_events.remove(ev)
                          
            renpy.notify("\n" + mod.name + " has been deactivated.")
            
            
                          
        def has_active_mod(self, name):
            if name in self.active_mods.keys():
                return True
            return False
            
        def list_free_girls(self):
            
            l = []
            
            for g in self.free_girls:
                l.append(g.name)
            
            return "Free girls: " + and_text(l)
        
        def get_available_locations(self):
            loc_list = []
            
            for d in district_dict.values():
                if d.rank <= district.rank:
                    loc_list += location_dict[d.name]
                    
            return loc_list
        
        def get_goal_description(self):
            
            text1 = and_text([g.get_description() for g in self.goals])
            
            if text1:
                return "{size=-1}To advance to the next chapter, " + text1 + ".{/size}"
            
            else:
                return "You are now in endless mode, enjoy continuing the game!"
        
        def get_task(self):
            if self.chapter > 1:
                return self.get_goal_description()
            else:
                return self.task
        
        def set_task(self, msg):
            self.task = msg
        
        def goals_reached(self):
            
            for goal in self.goals:
                if not goal.reached():
                    return False
            
            return True

        #<Chris12 AutoRepair>
        # If there could be new images, checks if all girls still have their portraits and profiles.
        # Does not perform any missing_girls business, use the help menu for that
        def update_files_timestamp(self):
            newFiles = None
            try :
                newFiles = self.__filesdict_timestamp != GirlFilesDict.get_timestamp()
            except :
                newFiles = True # For old savegame compatibility

            if newFiles:
                self.__filesdict_timestamp = GirlFilesDict.get_timestamp()
                for girl in (MC.girls + slavemarket.girls + game.free_girls + MC.escaped_girls + farm.girls):
                    girl.check_pictures()
        #</Chris12 AutoRepair>

    class Main(object): #Attributes: name, job, god, pictures, decisions, inventory, girl inventory, character
        
        """This class is for the main character."""
        
        def __init__(self):
            
            self.name = MC_name
            self.level = 1
            self.prestige = 0
            self.gold = starting_gold
            self.resources = defaultdict(int)
            self.last_collected = defaultdict(int)
            self.loan = None
            self.good = 0
            self.evil = 0
            self.neutral = 0
            self.girls = []
            self.escaped_girls = []
            self.trainers = []
            self.current_trainer = None
            self.items = []
            self.active_inv_filter = None
            self.effects = []
            self.effect_dict = defaultdict(list)
            self.equipped = []
            self.slots = MC_inventory_slots
            self.playerclass = None
            self.god = None
            self.noble = False
            self.active_spells = []
            self.known_spells = []
            self.skill_points = 0
            self.pics = []
            self.current_pic = "black"
        
        # Banking
        
        def take_loan(self, loan):
            if not self.loan:
                self.loan = loan
                self.gold += loan.amount
                return True
            return False
        
        def repay_loan(self):
            if self.loan:
                cost = self.loan.repay()
                if self.loan.amount <=0: # Loan repaid
                    calendar.set_alarm(calendar.time+1, Event(label = "loan_repaid"))              
                return "You paid " + str(round_int(cost)) + " gold to the bank for your loan."
            return False
            
        def repay_in_full(self):
            if self.loan:
                if self.gold >= self.loan.amount:
                    if renpy.call_screen("yes_no", "Are you sure you want to repay your loan in full for " + str(self.loan.amount) + " gold?"):
                        self.gold -= self.loan.amount
                        self.loan = None
                        return True
            return False
                
        
        ## Playerclass Init
        
        def set_playerclass(self, playerclass):
            
            self.playerclass = playerclass
            
            if playerclass == "Warrior":
                
                self.strength = 2
                self.spirit = 1
                self.charisma = 0
                self.speed = 3
                
            elif playerclass == "Wizard":
                
                self.strength = 0
                self.spirit = 2
                self.charisma = 1
                self.speed = 3
                
            elif playerclass == "Trader":
                
                self.strength = 1
                self.spirit = 0
                self.charisma = 2
                self.speed = 3
            
            self.reset_interactions()
            
            self.load_pics()
        
        ## Load pics
        
        def load_pics(self):
            
            # Loading files
            
            self.files = []
            
            for file in renpy.list_files():
            
                if file.startswith("MC/"):
                
                    file_parts = file.split("/")
                    file_name = file_parts[-1]
                
                    self.files.append(file)
            
            # Resetting pictures
            
            self.pics = []
            
            # Identifying image files
            
            imgfiles = [img for img in self.files if (img[-4:] in IMGFORMATS or img[-5:] in IMGFORMATS)]
            
            # Creating pictures
            
            for file in imgfiles:
                
                file_name = file.split("/")[-1]
                
                if self.playerclass.lower() in file_name.lower():
                
                    pic = Picture(file_name, file)
                    
                    self.pics.append(pic)
                    
            if not self.pics:
                raise AssertionError, "No picture found for the " + MC.playerclass + " class. Check the game/MC folder."
            else:
                self.current_pic = self.pics[0]
        
        def change_pic(self, direction):
            if direction == "next":
                self.current_pic = get_next(self.pics, self.current_pic, loop=True)
            elif direction == "previous":
                self.current_pic = get_previous(self.pics, self.current_pic, loop=True)
        
        ## Reset interactions (daily)
        
        def reset_interactions(self):
            self.interactions = self.get_speed()
            self.mana = self.get_spirit() * self.get_effect("boost", "mana") + self.get_effect("change", "mana")
        
        ## Resources
        
        def has_gold(self, amount):
        
            if self.gold >= amount:
                return True
            
            else:
                return False
        
        def gain_resource(self, resource="", number=1, message=True, random=False): # Where resource is the resource name
            
            if random:
                resource = rand_choice([r for r in build_resources if resource_dict[r].rank <= district.rank])
                if resource == "diamond": # can never get more than 1 diamond from a random city event
                    number=1
            
#            renpy.say("", "gaining " + str(resource))
            
            if resource == "prestige":
                self.prestige += number
            elif resource in ("gold", "money", "denar"):
                resource = "gold"
                self.gold += number
            else:
                self.resources[resource] += number
                
            if message:
                renpy.call("resource_gained", resource, number)
            else:
                renpy.notify("+" + str(number) + " " + resource)
        
        def collect_resource(self, resource): # Where resource is the resource name
            
#            renpy.say("", "collecting " + str(resource))
            
            if self.last_collected[resource] == calendar.time:
                return "KO"
            
            self.last_collected[resource] = calendar.time
            self.interactions -= 1
            
            res = resource_dict[resource]
            
            if res.rank == 2:
                nb = dice(3, 1 + int(self.get_stat(res.stat)/2)) # + self.get_effect("change", "basic resource extraction")
            elif res.rank == 3:
                nb = dice(2, 1 + int(self.get_stat(res.stat)/3)) # + self.get_effect("change", "avanced resource extraction")
            elif res.rank == 4:
                if renpy.random.random() <= 0.33: # 33% chance of finding diamond
                    nb = 1 # + self.get_effect("change", "diamond extraction")
                else:
                    renpy.call_screen("OK_screen", message="You failed to find anything.")
                    return
            else:
                return
            
            # The boost is generic (moon effect)
            nb += 1 * self.get_effect("boost", "resource extraction") + self.get_effect("change", resource + " extraction")
            
            self.gain_resource(resource, number=round_int(nb))
            
            return
                    
        def has_resource(self, resource, amount): # Where resource is the resource name
            if resource == "gold" and self.gold < amount:
                return False
            elif resource != "gold" and self.resources[resource] < amount:
                return False
            return True
        
        def exchange(self, source="gold", target="gold", target_amount=0):
            
            cost = get_exchange_rate(source, target) * target_amount
            
            if not self.has_resource(source, cost):
                renpy.say(market_girl, "I'm sorry, you just don't have enough " + source + " to pay for this.")
            
            else:
                renpy.play(resource_dict[target].sound, "sound")
                self.resources[source] -= cost
                self.resources[target] += amount
                
                
        ## Spells
        
        def has_spell(self, spl):
            
            for s in self.known_spells + self.active_spells:
                
                if s.name == spl.name:
                
                    return s
                
            return False
                
        def has_active_spell(self, spellname = None, spelltype = None):
            
            if spelltype:
                for s in self.active_spells:
                    if s.type == spelltype:
                        return True
            
            else:
                for s in self.active_spells:
                    if s.name == spellname:
                        return True
                
            return False
        
        def has_auto_spell(self, spelltype):
            
            for s in self.known_spells:
                if s.auto and s.type == spelltype:
                    return True
            
            return False
                
        
        def update_spells(self):
            
            for spl in spellbook[self.playerclass]:
                
                if spl.level <= self.level and not self.has_spell(spl):
                    
                    self.learn(spl)
                    
                elif spl.level > self.level and self.has_spell(spl): # Shouldn't be needed anymore
                    
                    self.unlearn(self.has_spell(spl))
                    renpy.pause(0.5)
                    
            self.known_spells.sort(key=lambda x: x.get_cost())
#            renpy.say("", "Known spells =" + str(len(self.known_spells)))
                    
        def learn(self, spl):
            if spl.type == "passive":
                self.active_spells.append(spl)
                self.add_effects(spl.effects)
                
                renpy.call_screen("OK_screen", title = spl.name, message = self.name + " has learnt a new talent.\n\n" + spl.description, pic = spl.pic, pic_size = "small")
            
            else:
                self.known_spells.append(spl)
                renpy.call_screen("OK_screen", title = spl.name, message = self.name + " has learnt a new spell.\n\n" + spl.description, pic = spl.pic, pic_size = "small")
            
            spl.auto = False

        def unlearn(self, spl): # Obsolete with 0.14
            
            if spl in self.active_spells:
                self.deactivate_spell(spl)
            
            self.known_spells.remove(spl)
            spl.auto = False

            renpy.notify("\n{color=[c_crimson]}" + self.name + " has forgotten " + spl.name + "{/color}")
                    
        
        def activate_spell(self, spl):
            
            if spl in self.active_spells:
                renpy.notify("\nThis spell is already active.")
                return False
                
            elif self.has_active_spell(spelltype = spl.type):
                renpy.notify("\nAnother " + spl.type + " spell is already active.")
                return False
                
            elif self.mana >= spl.cost:
                
                self.mana -= spl.cost
                self.active_spells.append(spl)
                self.add_effects(spl.effects)
                
                renpy.play(spl.sound, channel='sound2')
                renpy.notify("\n" + spl.name + " has been activated.")
                renpy.pause(0.5)
                    
                return True
                    
            else:
                
                renpy.notify("\nYou do not have enough mana to cast this spell.")
                
                return False
                
            return
        
        def autocast(self, spell, time):
            
            msg = ""
            
            if spell.auto == time and spell not in self.active_spells:
                if self.activate_spell(spell):
                    msg = "You have cast " + spell.name + "."
                    sound = s_spell
                else:
                    msg = "You failed to cast " + spell.name + "."
                    sound = s_fizzle
                
                renpy.play(sound, "sound")
                renpy.notify(msg)
                renpy.pause(0.5)
                
            return msg
            
                            
        def deactivate_spell(self, spl):
            
            self.active_spells.remove(spl)
            self.remove_effects(spl.effects)
                            
            renpy.notify(spl.name + " has expired")
            renpy.pause(0.5)
            
            return
            
        def reset_spells(self):
            for spell in [s for s in self.active_spells if s.duration == "turn"]:
                MC.deactivate_spell(spell)
        
        def toggle_auto_spell(self, spl, time="night"):
            
            if spl.auto == "night":
                spl.auto = "morning"
                
            elif spl.auto == "morning":
                spl.auto = False
            
            elif self.has_auto_spell(spl.type):
                renpy.notify("\nOnly one " + spl.type + " spell can be automated.")
                
            else:
                spl.auto = "night"
        
        
        ## Leveling
        
        def change_prestige(self, chg, apply_boost=True):
            if apply_boost:
                chg = chg * self.get_effect("boost", "prestige") + self.get_effect("change", "prestige")
            
            chg *= cheat_modifier["prestige"]
            self.prestige += chg
            
            if MC.ready_to_level():
                MC.level_up()
            
            return chg
        
        def ready_to_level(self):
            if self.level < 25:
                if self.prestige >= MC_xp_to_levelup[self.level]:
                    return True

            return False
            
        def level_up(self, forced = False):
            
            if (self.ready_to_level() or forced) and self.level < 25:
                self.level += 1
                self.skill_points += 1
                renpy.play(s_spell, "sound")
                self.update_spells()
                return True
                
            else:
                return False
        
        def change_stat(self, stat, nb, apply_boost=True, spillover=False):
            
            if apply_boost:
                nb = nb * self.get_effect("boost", stat + " gains") + self.get_effect("change", stat + " gains")
            
            if stat == "strength":
                if self.strength + nb <= 10:
                    self.strength += nb
                    return True
            
            elif stat == "spirit":
                if self.spirit + nb <= 10:
                    self.spirit += nb
                    return True
                    
            elif stat == "charisma":
                if self.charisma + nb <= 10:
                    self.charisma += nb
                    return True
                    
            elif stat == "speed":
                if self.speed + nb <= 10:
                    self.speed += nb
                    return True
                    
            return False
        
        def raise_stat(self, stat, nb): # Obsolete: rename to change_stat
            return self.change_stat(stat, nb)
        
        
        ## Get methods
        
        def get_stat(self, statname, raw=False):
            if statname.lower() in ("strength", "defense"):
                r = self.get_defense(raw=raw)
                
            elif statname.lower() == "spirit":
                r = self.get_spirit(raw=raw)
                
            elif statname.lower() == "charisma":
                r = self.get_charisma(raw=raw)
                
            elif statname.lower() == "speed":
                r = self.get_speed(raw=raw)
            
            return r
        
        def get_strength(self, fight=False, raw=False):
            return self.get_defense(fight, raw)
        
        
        def get_defense(self, fight=False, raw=False): # Raw affects item boosts but not spells
            
            defense = self.strength
            
            if not raw:
                defense += self.get_effect("change", "strength") + self.get_effect("change", "defense")
            
            if self.get_effect("special", "dragon form") > 0:
                
                defense = self.get_spirit()
            
            if fight and self.has_active_spell(spellname = "Summon bloodhound"):
                
                defense += dice(6)
            
            return defense
        
        
        def can_defend(self):
            if self.interactions > 0 or self.get_effect("special", "defender"):
                return True
            return False
        
        def get_spirit(self, raw=False):
            
            if raw:
                return self.spirit
            
            spirit = self.spirit

            spirit += self.get_effect("change", "spirit")
            
            return spirit
            
            
        def get_charisma(self, raw=False):

            char = self.charisma
                
            if not raw:
                char += self.get_effect("change", "charisma")
            
            if self.get_effect("special", "fairy form") > 0:
                
                char = self.get_spirit()
            
            return char
            
            
        def get_speed(self, raw=False):
            if raw:
                return self.speed
            
            spd = self.speed
                
            spd += self.get_effect("change", "speed")
            
            return spd            
        
        
        # Old, should be phased out in favor of challenges
        
        def test_defense(self, diff):
            
            your_score = self.get_defense(fight = True) + dice(6)
            enemy_score = diff + dice(6)
            
#            renpy.say("", "Rolled " + str(your_score) + "against " + str(enemy_score))
            
            if your_score >= enemy_score:
                return True
                
            else:
                return False
                
        def test_spirit(self, diff):
            
            if (self.get_spirit() + dice(6)) >= (diff + dice(6)):
                return True
                
            else:
                return False
                
        def test_charisma(self, diff):
            
            if (self.get_charisma() + dice(6)) >= (diff + dice(6)):
                return True
                
            else:
                return False
        
        def get_alignment(self):
            if self.good > self.evil + 2 and self.good > self.neutral:
                return "good"
            elif self.evil > self.good + 2 and self.evil > self.neutral:
                return "evil"
            else:
                return "neutral"
        
            
        ## Items
        
        
        def get_modifier(self, operation):
            
            mod = self.get_effect("boost", operation) * price_modifiers[operation]
            
            # Hard cap on the Haggler talent
            if (operation == "buy" and mod < 0.75) or (operation == "sell" and mod > 0.75):
                mod = 0.75
                
            return mod
            
        def buy(self, seller, obj, price):
            
            self.gold -= price
                
            self.take(seller, obj)
            
            renpy.block_rollback()
            
            
        def sell(self, buyer, obj, price = None):
            
            
#            if not price:
#                price = object.get_price("sell")
                
            self.gold += price
                
            self.give(buyer, obj)
            
            renpy.block_rollback()
            
            
        def take(self, giver, obj): ## I know, girls are not objects...
                
            if obj.type == "girl":
                acquire_girl(obj)
                try:
                    giver.girls.remove(obj)
                except:
                    pass
            else:
                self.items.append(obj)
                try:
                    giver.items.remove(obj)
                except:
                    pass
                
            renpy.restart_interaction()
            
            
        def give(self, taker, obj):
                
            if obj.type == "girl":
                if obj in self.girls:
                    self.girls.remove(obj)
                elif obj in farm.girls:
                    farm.girls.remove(obj)
                taker.girls.append(obj)
                
            elif taker:
                self.items.remove(obj)
                taker.items.append(obj)
            
            renpy.restart_interaction()
            
        def gift(self, taker, item):
            
            self.items.remove(item)
            
            renpy.say("", "You give " + taker.name + " {b}" + article(item.name) + "{/b}.")
            
            renpy.block_rollback()
            
            result = taker.receive_gift(item)
            
            return result
        
        def equip(self, item):
            
            for i in self.equipped:
                
                if i.type.slot == item.type.slot:
                    
                    self.unequip(i)
                    
            self.equipped.append(item)
            self.add_effects(item.effects)
            item.equipped = True
            
        def unequip(self, item):
            
            self.equipped.remove(item)
            self.remove_effects(item.effects)
            item.equipped = False
            
            
        def use_item(self, item):
            
            used = True
            r = ""
            
            if item.usage == "buff":
                
                self.add_effects(item.effects)
                
                calendar.set_alarm(calendar.time + item.charges, Event(label = "effect_expired", object = (self, item.effects)))
                
                self.items.remove(item)
                
            
            else:
                
                for e in item.effects:
                    
                    if e.type == "gain":
                                
                        if e.target in ("prestige"):
                            e.gain(self)
                            
                            if MC.ready_to_level():
                                MC.level_up()
                            
                        elif e.target in ("skills", "skill points"):
                            self.skill_points += e.value
                    
                            
            if used:
            
                r = item.use_me()

                if r == "used_up":
                    self.items.remove(item)
            
                renpy.block_rollback()
            
            return r
        
        def get_items(self, target="any", type="any", name="any", effect_type="any", effect_target="any"): # Where 'type' is a name, not an object
            
            items = []
            
            for it in self.items:
                if it.target == target or target == "any":
                    if it.type.name == type or type == "any":
                        if name in it.name or name == "any":
                            if effect_type != "any" or effect_target != "any":
                                if it.has_effect(effect_type, effect_target):
                                        items.append(it)
                            else:
                                items.append(it)
            return items
        
        ## Effects
        
        def get_effect(self, type, target):
            if type == "boost":
                return get_effect(self, type, target) * brothel.get_effect(type, target)
            else:
                return get_effect(self, type, target) + brothel.get_effect(type, target)
        
        def add_effects(self, effects, apply_boost=False, spillover=False, expires = False):
            return add_effects(self, effects, apply_boost=apply_boost, spillover=spillover, expires=expires)
            
        def remove_effects(self, effects):
            remove_effects(self, effects)

            
        ## Misc
            
        def cycle_trainers(self, reverse = False):
            
            index = self.trainers.index(self.current_trainer)
            
            if not reverse:
                index += 1
                if index > len(self.trainers)-1:
                    index = 0
            else:
                index -= 1
                if index < 0:
                    index = len(self.trainers)-1
            
            self.current_trainer = self.trainers[index]
            
            update_effects()
            
        def rand_say(self, sentences): # MC will say a random sentence in the list
            
            you(rand_choice(self.filter_say(sentences)))
            
            return
        
        def say(self, sentences): # Same as rand_say, just easier to manipulate
            return self.rand_say(sentences)
            
        def filter_say(self, sentences): # MC may say a random sentence in the list if and only if it checks out with his attributes
            
            d_list = []
            
            for it in sentences:
                
                if it.startswith("wr: ") or it.startswith("wa: "):
                    if self.playerclass == "Warrior":
                        d_list.append(it[4:])
                
                elif it.startswith("wz: ") or it.startswith("wi: "): 
                    if self.playerclass == "Wizard":
                        d_list.append(it[4:])
                    
                elif it.startswith("tr: "): 
                    if self.playerclass == "Trader":
                        d_list.append(it[4:])
                    
                elif it.startswith("gd: "):
                    if self.get_alignment() == "good":
                        d_list.append(it[4:])
                
                elif it.startswith("ne: "):
                    if self.get_alignment() == "neutral":
                        d_list.append(it[4:])
                
                elif it.startswith("ev: "):
                    if self.get_alignment() == "evil":
                        d_list.append(it[4:])
                        
                elif it.startswith("ar: "):
                    if self.god == "Arios":
                        d_list.append(it[4:])
                        
                elif it.startswith("sh: "):
                    if self.god == "Shalia":
                        d_list.append(it[4:])
                        
                elif it.startswith("ng: "):
                    if self.god == None:
                        d_list.append(it[4:])
                        
                else:
                    d_list.append(it)
            
            return d_list

    #### End of Main character class ####
    
    
    
    class NPC(): # Attributes: name, pictures, inventory, girl inventory, character
        
        """This class is for NPCs: story NPCs, shopkeeper and slave master, etc."""
        
        def __init__(self, name = "", char=None, defense=0, trainer_portrait = None, trainer_description = None, effects = None, item_types = "all"):
            self.name = name
            self.girls = []
            self.items = []
            self.love = 0
            self.hate = 0
            self.defense = defense
            self.met = False
            self.banged = False
            self.raped = False
            self.flags = defaultdict(bool)
            self.trainer_portrait = trainer_portrait
            self.trainer_description = trainer_description
            self.effects = make_tuple(effects, Effect)
            self.updated = False
            self.item_types = item_types
            if char:
                self.char = char
            else:
                self.char = Character(self.name)
        
        def get_defense(self, fight=False):
            return self.defense
        
#     class City(): # Note: Free girls should be transfered here from the game object
        
#         """This class is for the main city of Zan."""
        
#         def __init__():
#             self.effects = []
#             self.effect_dict = defaultdict(list)
            
        
    
    
    class District(): # Attributes: name, populations, brothels, competitors, vice, tooltip
        
        """This class covers the districts of the city, where each chapter of the game is played."""
        
        def __init__(self, name, chapter, rank, diff, pop, room = None, pic = None, locations = None, description = ""):
            
            self.name = name
            self.chapter = chapter
            self.rank = rank
            self.diff = diff
            self.pic = pic
            self.description = description
            self.quests_updated = -99
            if room:
                self.room = room
            else:
                self.room = []
            if locations:
                self.locations = locations
            else:
                self.locations = []
            
            self.items = defaultdict(list)
            
            self.no_reminder = -1 # storing the latest skipped day for the 'relocate/visit city' reminder
            
        def get_rand_pop(self):
            return weighted_choice([(p, p.weight) for p in all_populations])
            
        def get_pic(self, x, y):
            return im.Scale(self.pic, x, y)
        
        
    class Population():
        
        def __init__(self, name, pic, diff, range, rank=1, effects=[], weight=0, base_description = ""):
            
            self.name = name
            self.pic = Picture(pic, "UI/customers/" + pic)
            self.diff = diff
            self.range = range
            self.rank = rank
            self.effects = effects
            self.weight = weight
            self.description = "{b}" + self.name.capitalize() + "{/b} (difficulty: " + self.get_difficulty() + ")\n" + get_description(base_description, effects)
            
        def get_rand_name(self):
            return rand_choice(pop_name_dict[self.name])
                
        def get_effect(self, type, target):
            
            if type == "boost":
                r = 1 * brothel.get_effect(type, target)
            else:
                r = 0 + brothel.get_effect(type, target)
                
            for eff in self.effects:
                if eff.target.lower() == target.lower():
                    if renpy.random.random() <= eff.chance:
                        r += eff.value
            
            return r
        
        def get_pic(self, x, y):
            return self.pic.get(x, y)
        
        def get_difficulty(self):
            
            if self.diff <= 25:
                return event_color["good"] % str(self.diff)
            elif self.diff <= 50:
                return event_color["a little good"] % str(self.diff)
            elif self.diff <= 90:
                return event_color["average"] % str(self.diff)
            elif self.diff <= 120:
                return event_color["a little bad"] % str(self.diff)
            elif self.diff <= 170:
                return event_color["bad"] % str(self.diff)
            else:
                return event_color["fear"] % str(self.diff)
            
        
    class Customer():
        
        """This class is for individual customers interacting with the girls."""
        
        def __init__(self, pop):
            
            if pop:
                self.pop = pop
            else:
                self.pop = beggar
            
            self.rank = self.pop.rank
            self.name = article(self.pop.get_rand_name()).capitalize()
            
            self.randomize()
            
            self.reason = ""
            self.satisfaction = self.get_effect("change", "overall customer satisfaction")
            self.got_entertainment = None
            self.got_sex_act = None
            self.got_service = False
            self.entertained = False
            self.got_laid = False
            self.satisfied = False
            
            #<Chris Job Mod>
            if game.has_active_mod("chrisjobmod"):
                self.entertainment_score = unentertained_customer_score
            #</Chris Job Mod>
                    
        def randomize(self):
            
            self.preference = weighted_choice([("beauty", 25 + self.get_effect("change", "beauty preference")), ("body", 25 + self.get_effect("change", "body preference")), ("charm", 25 + self.get_effect("change", "charm preference")), ("refinement", 25 + self.get_effect("change", "refinement preference"))])
            self.wants_entertainment = weighted_choice([(j, customer_base_preference[j]*(1 + 0.5*game.customer_preference_weight[j]) + self.get_effect("change", j + " preference")) for j in all_jobs])
            self.wants_sex_act = weighted_choice([(a, customer_base_preference[a]*(1 + 0.5*game.customer_preference_weight[a]) + self.get_effect("change", a + " preference")) for a in all_sex_acts])
            self.fetish = rand_choice(trait_dict.keys())
            
            self.diff = self.pop.diff + dice(self.pop.range + 1) - 1 ## Varies up to +10 to +40 (maximum rank)
            self.defense = dice((district.rank) * 3) - 2
            
            if dice(100) <= 2 * self.get_effect("boost", "crazy") + self.get_effect("change", "crazy"):
                self.crazy = rand_choice(["arsonist", "rapist", "violent"])
            else:
                self.crazy = False

        def get_defense(self, fight = False):
            return self.defense * self.get_effect("boost", "customer defense") + self.get_effect("change", "customer defense")
        
        def get_effect(self, type, target):  
            return self.pop.get_effect(type, target)

        def get_description(self):
            return self.pop.description
        
        def choose_girl(self, girls, act): # For xxx interactions: returns a girl and a reason for choosing her.
            
            chosen = None
            best_score = 0
            
            # Raise errors
            
            if not girls:
                raise AssertionError, "Customer could not find girls to choose from."
            
            # Looks for the best girl
            
            for girl in girls:
                
                girl_score = 0
                reason = ""
                
                # 1. The customer looks for a particular fetish
                
                if girl.has_trait(self.fetish):
                    girl_score += 250
                    
                    if trait_dict[self.fetish].verb == "be":
                        reason = ":cust: came looking for a " + self.fetish + " girl. :pron: :verb: elated to meet :girl:."
                    elif trait_dict[self.fetish].verb == "be a":
                        reason = ":cust: came looking for a " + self.fetish + ". :pron: :verb: elated to meet :girl:."
                    elif trait_dict[self.fetish].verb == "have":
                        reason = ":cust: came looking for a girl with " + self.fetish + ". :pron: :verb: elated to meet :girl:."
                    elif trait_dict[self.fetish].verb == "have a":
                        reason = ":cust: came looking for a girl with a " + self.fetish + ". :pron: :verb: elated to meet :girl:."
                
                # 2. The customer looks for a particular stat
                
                girl_score += girl.get_stat(self.preference) # Customers are looking for one stat in particular
                
                if not reason:
                    reason = ":cust: wanted to meet " + gstats_descript[self.preference] + ". "
                    
                    if girl_score >= 50*self.rank:
                        reason += ":pron: :verb: elated to meet :girl:."
                    elif girl_score >= 35*self.rank:
                        reason += ":pron: :verb: pleased to meet :girl:."
                    elif girl_score >= 20*self.rank:
                        reason += ":pron: settled for :girl:." # Special case handled with a try / except clause within the perform method
                    else:
                        reason += ":pron: :verb: disappointed to meet :girl:."
                
                # 3. The customer looks for the best performer
                
                if act:
                    girl_score += girl.get_stat(self.wants_sex_act)
                    
                if girl.rank < self.rank: # Customers will favor girls with their rank or higher
                    girl_score -= 100
                
                # 4. Check score
                
                if girl_score > best_score:
                    chosen = girl
                    best_score = girl_score
                    self.reason = reason
                
            # Choose a random girl if no-one matches their tastes
            
            if not chosen:
                chosen = rand_choice(girls)
            if not self.reason:
                self.reason = ":cust: wanted to meet " + gstats_descript[self.preference] + ". :pron: couldn't find a suitable girl."
            
            return chosen

        def get_rep_bonus(self):
            bonus = 1 - self.rank
            
            if self.got_entertainment:
                bonus += 1
                if self.got_entertainment == self.wants_entertainment:
                    bonus += 1
            if self.got_sex_act:
                bonus += 1
                if self.got_sex_act == self.wants_sex_act:
                    bonus += 1
            
            return bonus
            
                
        
    ## The code for brothels and room is messy / redundant and should be cleaned up some time
        
    class Brothel(): #Attributes: name, rooms, reputation, advertisement, tooltip
        
        """This class is for available brothels that you can operate."""
        
        def __init__(self, level, upgrades, cost, max_rep, max_help):
            
            self.name = "The Rose Garden"
            self.rank = level # Level should be phased out and replaced with rank
            self.level = level
            self.cost = cost
            self.bedroom_type = room_dict[upgrades[0]]
            self.maxupgrade = upgrades[1]
            self.max_rep = max_rep
            self.max_help = max_help
            self.furniture = []
            self.rooms = common_room_dict
            self.effect_dict = defaultdict(list)
        
        # Set up Brothel
        
        def setup(self, name, furniture=[], current_building=None, started_building=0, master_bedroom_girls=[], free_room=[]): # Happens on a new chapter, and can be used anytime you need to reset the Brothel to 'factory settings'
                        
            self.pic = im.Scale("brothels/" + brothel_pics[self.level], config.screen_width, int(config.screen_height*0.8))
            self.pic_index = self.level
            # Unlocks pics for gallery
            unlock_pic("brothels/" + brothel_pics[game.chapter])
            unlock_pic(self.bedroom_type.pic_path)

            self.rep = 1
            self.advertising = 0
            
            self.maintenance = 0
            self.maint_value = 1
            self.dirt = 0
            
            self.security = 0
            self.threat = 0 # Threat builds up over time and causes security events to proc
            self.alert_level = 1 # Alert level determines the type of security events that can proc
                                    
            if free_room == "free":
                self.free_room = True
            
            else:
                self.free_room = False
                for room in free_room:
                    self.rooms[room].buy(forced=True)
                        
            self.bedrooms = bro_capacity[game.chapter][0]
                        
            self.magic_shield = False
            
            self.name = name
            self.furniture = furniture
            self.current_building = current_building
            self.started_building = started_building
            
            self.master_bedroom = master_bedrooms[self.rank-1]
            self.master_bedroom.girls = master_bedroom_girls
            
        
        def get_maxbedrooms(self): # Replaces the old 'maxbedrooms' property
            return bro_capacity[game.chapter][1]
        
        # Master bedroom
        
        def upgrade_master_bedroom(self, target_level="auto"):
            if target_level == "auto":
                target_level = self.master_bedroom.level + 1
            
            if MC.has_gold(master_bedrooms[target_level].cost):
                if renpy.call_screen("yes_no", "Are you sure you want to upgrade your room for " + str(master_bedrooms[target_level].cost) + " gold?"):
                    renpy.play(s_gold, "sound")
                    MC.gold -= master_bedrooms[target_level].cost
                    temp_girls = self.master_bedroom.girls
                    self.master_bedroom.girls = []
                    
                    self.master_bedroom = master_bedrooms[target_level]
                    
                    self.master_bedroom.girls = temp_girls
                    
            else:
                renpy.say(sill, "Sorry Master, you do not have enough money.")
                    
        
        def can_have(self, job):
            if job == "farm":
                return False
            elif job not in all_jobs:
                return True
            elif self.rooms[job_room_dict[job.lower()]].level > 0:
                return True
            else:
                return False
        
        # Furniture
        
        def buy_furniture(self, furn):
            
            sounds = []
            
            if self.current_building:
                renpy.say(carpenter, "Sorry boss, I still have work to do on that " + self.current_building.name + " you ordered.")
                return False
            
            elif not furn.can_build():
                if furn.built:
                    renpy.say(carpenter, "You've got it already. I don't think you need a second one.")
                renpy.say(carpenter, "Cannot build that for now, I'm 'fraid. You'd need a larger place.")
                return False
            
            for resource, amount in furn.cost:
                if not MC.has_resource(resource, amount):
                    renpy.say(carpenter, "Look, boss, you gotta have the right amount of resources before I can start the job.")
                    break
            else:
                renpy.say(carpenter, "A'right, looks like you've got the goods. Hand them over, and I'll get started on that " + furn.name + " right away.")
                if renpy.call_screen("yes_no", "Are you sure you want to build a " + furn.name + " for " + furn.describe_cost() + "?"):
                    for resource, amount in furn.cost:
                        if resource == "gold":
                            sounds.append(s_gold)
                            MC.gold -= amount
                        else:
                            MC.resources[resource] -= amount
                            sounds.append(resource_dict[resource].sound)
                    renpy.play(rand_choice(sounds), "sound")
                    renpy.block_rollback()
                    furn.start_building()
                    return True
            return False
                
        def activate_furniture(self, furn):
            return furn.activate()
        
        def deactivate_furniture(self, furn):
            return furn.deactivate()
            
        def toggle_furniture(self, furn):
            return furn.toggle()
        
        def destroy_furniture(self, furn):
            furn.destroy()
                
        
        # Effects
        
        def get_effect(self, type, target):

            return get_effect(self, type, target)
        
        def update_effects(self):
            
            update_effects()            

        
        ## Important: Security, advertising and maintenance change effects are applied BEFORE boost effects
        
        def get_security(self):
            
            return ((self.security + self.get_effect("change", "security")) * self.get_effect("boost", "security"))

            
        
        def get_advertising(self):
            
            return (self.advertising + self.get_effect("change", "advertising")) * self.get_effect("set", "advertising value") * self.get_effect("boost", "advertising")
            
            
        def get_maintenance(self):
        
            return (self.maintenance + self.get_effect("change", "maintenance")) * self.get_effect("boost", "maintenance")
        
        
        ## Brothel Threat
        
        def get_gold_threat(self):
            
            gt = (MC.gold - self.get_effect("special", "safe")) / gold_threat_amount[district.rank]
            
            if gt > gold_threat_max[game.chapter]:
                return gold_threat_max[game.chapter]
            else:
                return gt
        
        def get_threat(self): # Threat is influenced by district rank, gold (up to a maximum) and the number of working girls
            
            threat = district.rank + self.get_gold_threat() + sum(girl.rank/2.0 for girl in MC.girls if girl.works_today())
            
            return threat * self.get_effect("boost", "threat") + self.get_effect("change", "threat")
        
        def get_risk(self):
#            threat = self.get_threat()
#            defense = self.get_security() + MC.get_defense() MC defense no longer helps with security, but it matters during actual events

            return self.get_threat() - self.get_security()
        
        def estimate_threat_level(self, contrast=False):
            
            risk = self.get_risk()
            
            if risk >= 10:
                
                level = event_color["bad"] % "{b}very high{/b}"
                
            elif risk >= 5:
                if contrast:
                    level = event_color["a little bad contrast"] % "{b}high{/b}"
                else:
                    level = event_color["a little bad"] % "{b}high{/b}"
                
            elif risk <= -10:
                
                level = event_color["good"] % "{b}very low{/b}"
                
            elif risk <= -5:
                
                level = event_color["a little good"] % "{b}low{/b}"
                
            else:
                if contrast:
                    level = event_color["normal contrast"] % "{b}normal{/b}"
                else:
                    level = event_color["normal"] % "{b}normal{/b}"
                
            return level
        
        def threat_build_up(self): # Builds up threat every turn depending on active security, with a minimum of 1. Returns True if security event can proc.
        
            if self.get_risk() < 1:
                self.threat += 1
            else:
                self.threat += self.get_risk()
                
            if self.threat >= 50 and self.alert_level == 3:
                return True
            elif self.threat >= 25 and self.alert_level == 2:
                return True
            elif self.threat >= 10 and self.alert_level == 1:
                return True
            
            return False
        
        # Advertising
        
        def estimate_customers(self):
            
            future_rep = self.rep + self.get_advertising()
            
            if future_rep > self.max_rep:
                future_rep = self.max_rep
            elif future_rep < 1:
                future_rep = 1

            return count_customers(future_rep, [girl for girl in MC.girls if girl.works_today()])
        
        
        def get_ASM_report(self):
            msg = ""
            
            cust, extra = brothel.estimate_customers()
            
            msg += "Advertising report: {b}" + str(cust + extra) + " customers{/b} are expected to come tonight"
            
            if extra:
                msg += " (including " + str(extra) + " from girl or brothel effects)"
                                
            msg += ".\nSecurity report: The threat to your brothel is " + brothel.estimate_threat_level(contrast=True) + "."
            
            clean = brothel.get_cleanliness()
            
            msg += "\nMaintenance report: " + maintenance_desc[clean]
            
            return msg
            
        
        def change_rep(self, value):
            
            value = value * self.get_effect("boost", "brothel rep gains") + self.get_effect("change", "brothel rep gains")
          
            if self.rep + value < 1:
                chg = 1 - self.rep
                self.rep = 1
                
            elif self.rep + value > self.max_rep:
                chg = self.max_rep - self.rep
                self.rep = self.max_rep
                
            else:
                chg = value
                self.rep += value
            
            return chg
        
        # ROOMS
        
        def upgrade_bedrooms(self):
            
            price = self.get_room_upgrade_price(self.bedrooms)
            
            text1 = "Do you really want to upgrade the bedrooms for " + str(price) + " gold?"
            
            if self.bedroom_type.level < self.maxupgrade:
            
                if renpy.call_screen("yes_no", text1):
                    
                    if MC.has_gold(price):
                        
                        MC.gold -= price
                        
                        self.bedroom_type = room_dict[self.bedroom_type.level+1]
                        unlock_pic(self.bedroom_type.pic_path)
                        
                        renpy.block_rollback()
                        
                        renpy.restart_interaction()
                        
                    else:
                        renpy.say(narrator, "You don't have enought money.")
                        
            else:
                renpy.say(sill, "You cannot upgrade this brothel's rooms further.")
                
        def get_mood_modifier(self, rank): #Increases with bedroom type: Girls score higher with customers and their mood improves
            
            mood_modifier = self.bedroom_type.level + self.get_effect("change", "mood modifier") - (rank * 2)
            
            return mood_modifier
            
            
        def get_room_price(self, room = "bedroom"):
            
            if room == "bedroom":
                price = (100 * self.bedrooms)
                
                for lvl in range(self.bedroom_type.level):
                    
                    price += 50 * lvl
                    
                price *= district.rank #! Increased bedroom price
                
            else:
                
                if brothel_firstvisit or self.free_room:
                    price = 0
                    
                else:
                    price = self.rooms[room].get_price()
                    
            return price
        
        def get_room_upgrade_price(self, nb = 1):
        
            # This is the cost per bedroom
            price = 50 * self.bedroom_type.level * district.rank #! Increased room upgrade price
        
            return price * nb
        
            
        def add_room(self, room = "bedroom", forced=False):
            
            if room == "bedroom":
                
                price = self.get_room_price()
                
                text1 = "Do you want to buy a new bedroom for " + str(price) + " gold?"
                
                if self.bedrooms < self.get_maxbedrooms():
                
                    if renpy.call_screen("yes_no", text1):
                        
                        if MC.has_gold(price):
                            
                            MC.gold -= price
                        
                            self.bedrooms += 1
                            
                            renpy.block_rollback()
                        
                            renpy.restart_interaction()
                            
                            return True
                        
                        else:
                            renpy.say(narrator, "You don't have enought money.")
                            return False
                else:
                    renpy.say(sill, "You already have the maximum number of bedrooms for this brothel.")
                    return False
                
            else:
                self.rooms[room].buy(forced)
        
        def upgrade_room(self, room):
            self.rooms[room].upgrade()
        
        def get_common_rooms(self):
            return [room for room in self.rooms.values() if room.level >0]
        
        def get_bedroom_pic(self, x, y):
            return self.bedroom_type.get_pic(x, y)
        
        def get_room_pic(self, type, x, y):
            if type == "bedroom":
                return self.get_bedroom_pic(x, y)
            elif type.lower() == "tavern":
                return tavern.get_pic(x, y)
            elif type.lower() in ("club", "strip club"):
                return club.get_pic(x, y)
            elif type.lower() == "onsen":
                return onsen.get_pic(x, y)
            elif type.lower() == "okiya":
                return okiya.get_pic(x, y)
            
        def get_adv_cost(self):
            
            return self.advertising * helper_cost[district.rank]
        
        def get_sec_cost(self):
            
            return self.security * helper_cost[district.rank]
        
        def get_maintenance_cost(self):
            
            return self.maintenance * helper_cost[district.rank]
            
        def change_dirt(self, nb):
            
            boost = self.get_effect("boost", "dirt")
            
            boost = reverse_if(boost, nb)
            
            nb *= boost
            
            if self.dirt + nb < 0:
                nb = -self.dirt
                self.dirt = 0
            elif self.dirt + nb > 1000:
                nb = 1000 - self.dirt
                self.dirt = 1000
            else:
                self.dirt += nb
                
            return nb
        
        def get_cleanliness(self):
            
            if self.dirt < 10:
                return "clean"
            elif self.dirt < 25*district.rank:
                return "clean enough"
            elif self.dirt < 50*district.rank:
                return "dusty"
            elif self.dirt < 75*district.rank:
                return "dirty"
            elif self.dirt < 100*district.rank:
                return "disgusting"
            else:
                return "fire"
        
        def clean_up(self):
            
            price = self.get_clean_up_cost()
            
            if MC.has_gold(price):

                MC.gold -= price
                self.dirt = 0
                renpy.block_rollback()
                
                return True

            else:
                renpy.say(narrator, "You don't have enought money.")
                
                return False
        
        
        def get_clean_up_cost(self):
            
            return round_int(1.2 * (helper_cost[district.rank]*brothel.dirt/brothel.maint_value))
        
            
        def has_room(self, room = "any"):
            
            if room == "bedroom":
                return True
            
            elif room != "any":
            
                if self.rooms[room].level > 0:
                    return True
                    
                else:
                    return False
                    
            else:
                if self.get_common_rooms():
                    return True
                return False
        
        
        def cycle_pic(self, reverse = False):
            
            if not reverse:
                self.pic_index += 1
                if self.pic_index > 7:
                    self.pic_index = 1
            else:
                self.pic_index -= 1
                if self.pic_index < 1:
                    self.pic_index = 7
            
            renpy.notify("\nCycling..." + str(self.pic_index))
            
            self.pic = im.Scale("brothels/" + brothel_pics[self.pic_index], config.screen_width, int(config.screen_height*0.8))
            
            renpy.jump("brothel")
        
        def get_cust_count(self, breakdown=False):
            
            base = self.rep
            total_cust_value = 0
            
            cust_nb = 0
            adv = self.get_advertising()
            girl_bonus = 0
            
            for girl in [g for g in MC.girls if g.works_today()]:
                if girl.has_perk("Ponygirl"):
                    girl_bonus += 2 * girl.rank
                    
#                 if girl.has_perk("Exhibitionist"):
#                     if girl.get_effect("special", "exhib") != 1.0:
#                         girl_bonus += 1 * girl.rank
            
            
            while total_cust_value < base: # Each new customer "costs" 1 more point of rep times currents rank
                cust_nb += 1
                total_cust_value += cust_nb*district.rank
            
            base_cust_nb = cust_nb
            
            while total_cust_value < base + adv: # Each new customer "costs" 1 more point of rep
                cust_nb += 1
                total_cust_value += cust_nb
                
            adv_cust_nb = (cust_nb - base_cust_nb)*district.rank
            
            while total_cust_value < base + adv + girl_bonus:
                cust_nb += 1
                total_cust_value += cust_nb*district.rank
                
#            cust_nb += extra_cust
            
            girl_cust_nb = cust_nb - adv_cust_nb - base_cust_nb
            
            if breakdown:
                return cust_nb, base_cust_nb, adv_cust_nb, girl_cust_nb
            else:
                return cust_nb
            
            
        def get_cust_count_description(self):
            
            cust_nb, base_cust_nb, adv_cust_nb, girl_cust_nb = self.get_cust_count(True)

            des = str(cust_nb) + " customers are expected to come to the brothel tonight."
            
            if cust_nb != base_cust_nb:
                des += " ("
                
                if adv_cust_nb:
                    des += event_color["good"] % ("+" + str(adv_cust_nb) + " from advertising")
                    if girl_cust_nb:
                        des += ", "
                if girl_cust_nb:
                    des += event_color["good"] % ("+" + str(girl_cust_nb) + " from girls")
                    
                des += ")"
            
            return des + "."
        
        
        
#    class Competitor(): #Attributes: name, base_rep, girl_nb, av_girl_rep, tooltip
        
#        """This class is for competing brothels in the district."""
        
        
    class Calendar(object):
        
        """This class manages the game calendar, day/night cycle, and updates."""
        
        def __init__(self):
            
            self.time = 1
            self.day = 1
            self.month = 1
            self.year = 1
            self.moon = None
            self.alarms = {} # Events that proc on a particular day
            self.discounted = [] # Cheaper resources
            self.scarce = [] # More expensive resources
                    
        def get_date(self, time):
            
            year_time = 1 + (time-1) % (28*12)
            
            month = 1 + (year_time-1) // 28
            
            day = 1 + (year_time-1) % 28
            
            return str(month) + "/" + str(day)
        
        def newday(self, number = 1):
            
            for i in range(number):
            
                self.time +=1
                
                if self.day != 28:
                    self.day +=1
                
                else:
                    self.day = 1
                    
                    if self.month != 12:
                        self.month += 1
                    
                    else:
                        self.month = 1
                        self.year += 1
                        
            if self.time % 7 == 1 :
                self.updates()
            
            
        def updates(self): #This is the place where weekly updates are handled
            
            weekly_updates()
            
            return
  
  
        def get_weekday(self):
            
            wd = self.day % 7 -1
            
            return weekdays[wd]
        
        
        def get_discount(self, source, target): # Where source, target are "gold" or a Resource object
            
            if source == "gold" and target.name in self.discounted:
                return 0.75
            elif source == "gold" and target.name in self.scarce:
                return 1.25
            elif target == "gold" and source.name in self.discounted:
                return 0.75
            elif target == "gold" and source.name in self.scarce:
                return 1.25
            else:
                rate = 1.0
                
                if source != "gold":
                    if source.name in self.discounted:
    #                    renpy.notify("found " + source + "in discounts 1")
                        rate *= 3/4.0
                    if source.name in self.scarce:
    #                    renpy.notify("found " + source + "in scarce 1")
                        rate *= 3/2.0
                if target != "gold":
                    if target.name in self.discounted:
    #                    renpy.notify("found " + source + "in discounts 2")
                        rate *= 3/2.0
                    if target.name in self.scarce:
    #                    renpy.notify("found " + source + "in scarce 2")
                        rate *= 3/4.0
                    
                return rate
            
        def set_alarm(self, time, event):
            
            if isinstance(event, StoryEvent):
                daily_events.append(event)
                event.date = time
            
            else:
                if time in self.alarms:
                    self.alarms[time].append(event)
                
                else:
                    self.alarms[time] = [event, ]
                
            if debug_mode:
                renpy.notify("\nAlarm set for " + str(time))
                
                
        def play_alarms(self, time = None):
            
            event_list = []
            
            if not time:
                time = self.time
            
            # Event objects
            
            if time in self.alarms:
                
                event_list = self.alarms[time]
                
                del self.alarms[time]
            
            # StoryEvent objects
            
            for ev in daily_events:
                if ev.happens(type="day"):
                    event_list.append(ev)
            
            if event_list:
            
                event_list.sort(key=lambda x: x.order)
            
                renpy.call("display_events", event_list)
            
        def get_season(self):
            
            if self.month <= 3:
                
                return "winter"
                
            elif self.month <= 6:
                
                return "spring"
                
            elif self.month <= 9:
                
                return "summer"
                
            elif self.month <= 12:
                
                return "fall"

        
                        

        
    class Log(object): # One log object is created each day to track various events
        
        def __init__(self, time):
            
            self.time = time
            self.cust = 0
            self.cust_served = 0
            self.gold_made = 0
            self.upkeep = 0
            self.costs = 0
            self.net = 0
            self.report = "{color=[c_white]}" + calendar.get_weekday() + ", Y" + str(calendar.year) + " M" + str(calendar.month) + " D" + str(calendar.day) + "{/color}\n"
            self.events = []
            self.changes = ""
            self.track_dict = defaultdict(int)
        
        def track(self, key, value):
            self.track_dict[key] += value
        
        def get_day_report(self):
            
#            net = round_int(self.gold_made - self.upkeep - self.costs)
            
            if self.net >= 0:
                msg = "You made " + event_color["good"] % (str(round_int(self.net)) + " gold") + " last night.\n"
            else:
                msg = "You lost " + event_color["bad"] % (str(round_int(self.net)) + " gold") + " last night.\n"
            
            msg += "{size=-2}" + "- Gold made: + " + event_color["good"] % str(round_int(self.gold_made)) + "\n"
            msg += "- Girls upkeep: - " + event_color["bad"] % str(round_int(self.upkeep)) + "\n"
            msg += "- Brothel costs: - " + event_color["bad"] % str(round_int(self.costs)) + "\n{/size}\n"
            
            msg += str(self.cust) + " customer" + plural(self.cust) + " came to the brothel.\n"
            msg += "{size=-2}" + "- Customer served (job): " + event_color["good"] % str(self.check("served")) + "/" + str(self.cust) + "\n"
            msg += "- Customer entertained (job): " + event_color["good"] % str(self.check("entertained")) + "/" + str(self.check("served")) + "\n"
            msg += "- Customer served (sex): " + event_color["good"] % str(self.check("laid")) + "/" + str(self.cust) + "\n"
            msg += "- Customer satisfied (sex): " + event_color["good"] % str(self.check("satisfied")) + "/" + str(self.check("laid")) + "\n{/size}\n"
            
            msg += str(self.check("work_days")) + " girl" + plural(self.check("work_days")) +" worked in the brothel. " 
            
            if self.check("strike_days"):
                msg += event_color["bad"] % (str(self.check("strike_days")) + " girl" + plural(self.check("strike_days")) + " went on strike. ")
            if self.check("run_away"):
                    msg += event_color["bad"] % (str(self.check("run_away")) + " girl" + plural(self.check("run_away")) + " ran away from the brothel.")
                    
            msg += "\n{size=-2}" + "- Waitress: " + event_color["good"] % str(self.check("waitress_days")) + "\n"
            msg += "- Dancer: " + event_color["good"] % str(self.check("dancer_days")) + "\n"
            msg += "- Masseuse: " + event_color["good"] % str(self.check("masseuse_days")) + "\n"
            msg += "- Geisha: " + event_color["good"] % str(self.check("geisha_days")) + "\n"
            msg += "- Whore: " + event_color["good"] % str(self.check("whore_days")) + "\n{/size}\n"
            
            if self.check("rest_days") > 1:
                msg += str(self.check("rest_days")) + " girls were resting at the brothel. "
            elif self.check("rest_days") > 0:
                msg += str(self.check("rest_days")) + " girl was resting at the brothel. "
            
            if self.check("hurt_days") > 1:
                msg += event_color["bad"] % (str(self.check("hurt_days")) + " girls were hurt while working. ")
            elif self.check("hurt_days") > 0:
                msg += event_color["bad"] % (str(self.check("hurt_days")) + " girl was hurt while working. ")
                
            if self.check("exhausted"):
                msg += event_color["bad"] % (str(self.check("exhausted")) + " girl" + plural(self.check("exhausted")) + " became exhausted while working. ")
                   
            msg += "\n\n"
            
            if farm.active:
                msg += str(self.check("farm_days")) + " girls were at the farm last night. "
                
                if self.check("farm_resisted_training"):
                    msg += event_color["bad"] % (str(self.check("farm_resisted_training")) + " girl" + plural(self.check("farm_resisted_training")) + " resisted training. ")
                
                if self.check("farm_run_away"):
                    msg += event_color["bad"] % (str(self.check("farm_run_away")) + " girl" + plural(self.check("farm_run_away")) + " ran away from the farm.")
                
                if self.check("farm_hurt"):
                    msg += event_color["bad"] % (str(self.check("farm_hurt")) + " girl" + plural(self.check("farm_hurt")) + " got hurt while resisting.")
                    
                if self.check("minion_hurt"):
                    msg += event_color["bad"] % (str(self.check("minion_hurt")) + " minion" + plural(self.check("minion_hurt")) + " got hurt in the fighting.")
                
                msg += "\n{size=-2}" + "- In training: " + event_color["good"] % str(self.check("farm_training_days")) + "\n"
                msg += "- In holding: " + event_color["good"] % str(self.check("farm_holding_days")) + "\n{/size}\n"
                
                if self.check("farm_rest_days") > 1:
                    msg += str(self.check("farm_rest_days")) + " girls were resting at the farm. "
                elif self.check("farm_rest_days") > 0:
                    msg += str(self.check("farm_rest_days")) + " girl was resting at the farm. "
            
            return msg
            
        def get_tonight_report(self):
            pass
        
        def check(self, key):
            return self.track_dict[key]
                    
        def add_report(self, text):
            
            self.report += "\n" + text
    
        def add_event(self, event):
            
            self.events.append(event)
            
        def show_events(self):
            
            renpy.block_rollback()
            
            for event in self.events:
                
                renpy.checkpoint()
                
                if event == self.events[-1]:
                    
                    renpy.choice_for_skipping()
                
                event.show_night()

            
        
        



### SECONDARY CLASSES ##
    
    class Event(object):
        
        """This class covers 2 kinds of events: Night events (run during working hours) and Day events (run when returning to the main screen). Day events should be phased out and replaced by Story Events"""
        
        def __init__(self, pic = None, background = None, char = None, text = "", changes = "", sound = None, with_st = None, type="Normal", label = None, object = None, order = 0, weight = 1):
            
            self.pic = pic
            self.background = background
            self.char = char
            self.text = text
            self.sound = sound
            self.with_st = with_st
            self.type = type
            self.changes = changes #+ "\n" + type
            self.label = label #For day events
            self.object = object #For day events
            self.order = order #For day events
        
        def show_night(self):
            
#            renpy.show_screen("night", self.pic)

            log.changes = self.changes
            
            if self.sound:
                renpy.play(self.sound, "sound")
            
            renpy.say(self.char, self.text)
        
        def happens(self):
            return True
        
        def play(self):
            
            if self.label:
                
                if self.object:
                    renpy.call(self.label, self.object)
                    
                else:
                    renpy.call(self.label)
                
            else:
                
                if self.sound:
                    renpy.play(self.sound, "sound")
            
                renpy.say(self.char, self.text)
            
            
    class StoryEvent():
        
        def __init__(self, label, chapter=0, date=0, chance = 1.0, type="any", location = None, min_gold = -99999, condition = None, not_condition = None, condition_func=None, call_args=[], once = True, AP_cost = 1, order = 0, weight = 1, room = None):
            
            self.label = label
            self.chapter = chapter
            self.date = date
            self.chance = chance
            self.type = type # Type can be: "any" (plays anytime), "city", "day" (plays on main screen), "night" (plays upon ending day), "morning" (plays after night events)
            self.location = location
            self.min_gold = min_gold
            self.condition = condition
            self.not_condition = not_condition
            self.condition_func = condition_func # For complex conditions, this label is called and must 
            self.room = room
            self.call_args = call_args # call_args must be a list
            self.once = once
            self.AP_cost = AP_cost
            self.order = order
#            self.weight = weight

            self.happened = False
            self.mod = None
            
        def happens(self, type="any"): #Tests if happened, current chapter, chance of happening, location and custom story flags (optional)
            
#            renpy.say("checking", self.label)
            
            try: # Avoid breaking 0.15 saved games
                if self.condition_func: # A custom function that must return 'True' or 'False'
                    if not self.condition_func():
                        return False
            except:
                pass
            
            if self.mod:
                if self.mod not in game.active_mods.values():
                    return False
            
            if type != "any" and self.type not in ("any", type):
                return False
    
            if self.happened and self.once:
                return False
            
            if game.chapter < self.chapter:
                return False
            
            if calendar.time < self.date:
                return False
            
            if self.min_gold > MC.gold:
                return False
            
            if renpy.random.random() > self.chance:
                return False
            
            if self.location:
                if self.location.lower() != selected_location.name.lower():
                    return False            
            
            if self.condition:
                if self.mod:
                    if not self.mod.flags[self.condition]:
                        return False
                    
                elif not story_flags[self.condition]:
                    return False
                    
            if self.not_condition:
                if self.mod:
                    if self.mod.flags[self.not_condition]:
                        return False
                elif story_flags[self.not_condition]:
                    return False
            
#            if self.room:
#                if not brothel.has_room(self.room):
#                    return False
            
#            renpy.say("", self.label + "HAPPENS")
            
            return True
            
        def play(self):
            
            if self.once:
                r = story_remove_event(self.label)
                
                if not r:
                    if self.type == "city" and self in city_events:
                        city_events.remove(self)
                    elif self.type != "city" and self in daily_events:
                        daily_events.remove(self)
                                    
            renpy.call(self.label, *self.call_args)
                    
            # Renpy probably doesn't reach those two command lines, so I have included them within the display_events label instead
            self.happened = True
            
            story_flags[self.label] = True
            
            return
    
    
    class Stat(object):
        
        """This class is for stats (skills with a value that can be changed)."""
        
        def __init__(self, name, type, parent, weight=0):
            
            self.name = name
            self.type = type
            self.parent = parent # For now, parent can be a Girl object only
            self.statmax = self.get_statmax()
            self.init_value(weight)
            self.lastvalue = self.value
            
        def get_statmax(self):
            statmax = self.parent.rank * 50
            statmax += self.parent.get_effect("change", self.name.lower(), change_cap=True) + self.parent.get_effect("change", self.name.lower() + " max") + self.parent.get_effect("change", self.name + "all skill max")
            
#            if self.parent in MC.girls + farm.girls:
#                statmax += brothel.get_effect("change", self.name.lower()) + brothel.get_effect("change", "all skills") + brothel.get_effect("change", self.name.lower() + " max") + brothel.get_effect("change", self.name + "all skill max")
            
            return statmax
            
        def init_value(self, weight):
            
            if self.type == "sex":
                pref = self.parent.get_preference(self.name.lower())
                
                if pref == "refuses":
                    a = 0
                    b = -5
                elif pref in ("very reluctant", "reluctant"):
                    a = 1.5
                    b = 3.5
                elif pref in ("a little reluctant", "indifferent"):
                    a = 5
                    b = 10
                elif pref in ("a little interested", "interested"):
                    a = 7.5
                    b = 17.5
                elif pref in ("very interested", "fascinated"):
                    a = 12.5
                    b = 22.5                
            else:
                
                # Loads the dice if specified in init file
                
                if weight == 5:
                    d = 100
                elif weight == 4:
                    d = 90
                elif weight == 3:
                    d = 50
                elif weight == 2:
                    d = 20
                elif weight == 1:
                    d = 0
                else:
                    d = dice(100)
                
                if self.type == "sex":
                    
                    a = 0
                    b = -5
                
                elif d > 98: # Superb skill
                    
                    a = 12.5
                    b = 22.5
                
                elif d > 85: # High skill
                    
                    a = 7.5
                    b = 17.5
                    
                elif d > 35: # Average skill
                    
                    a = 5
                    b = 10
                    
                elif d > 10: # Low skill
                    
                    a = 1.5
                    b = 3.5
                    
                else: # Terrible skill
                    
                    a = 0
                    b = -5
                
            self.value = dice(5 + 5 * district.rank) + a * district.rank + b
            
            if self.value < 0:
                self.value = 0
                    
                
        
        def change(self, chg, max = 250):
            
            if self.value + chg < 0:
                
                r = -self.value
                
                self.value = 0
                
                return r
                
            elif self.value >= max:
                
                r = 0
                
                self.value = max
                
                return r
                    
            elif self.value + chg >= max:
                
                r = max - self.value
                
                self.value = max
                
                return r
            
            else:
                self.value += chg
                
                return chg
        
        
    class Trait(object):
        
        """This class is for traits (skills with special effects that are either on or off)."""
        
        # eff1, eff2, eff3 are kept for backwards compatibility, until I clean up the code
        
        def __init__(self, name, verb, eff1 = None, eff2 = None, eff3 = None, effects = [], opposite = [], archetype = None, base_description = ""):
            
            self.name = name
            self.verb = verb
            
            self.effects = []
            
            if effects:
                self.effects = effects
            if eff1:
                self.effects.append(eff1)
            if eff2:
                self.effects.append(eff2)
            if eff3:
                self.effects.append(eff3)
                    
            self.opposite = make_tuple(opposite)
            self.archetype = archetype
            
            self.base_description = base_description
            
#            renpy.say(self.name, str(len(self.effects)))
        
        def get_past_tense(self):
            
            if self.verb.startswith("be"):
                text1 = "was"
            elif self.verb.startswith("have"):
                text1 = "had"
        
            return self.add_article(text1) + " {b}" + self.name.lower() + "{/b}"
        
        def add_article(self, mytext):
            if self.verb.endswith("a"):
                mytext += " a"
            elif self.verb.endswith("an"):
                mytext += " an"
            
            return mytext
        
        def get_description(self, context=None, short=False):
            
            if short:
                return get_description("", self.effects)
            else:
                des = get_description(self.base_description, self.effects)
            
                if context == "slavemarket":
                    if self.archetype:
                        des += "\nUnlocks {b}" + self.archetype + "{/b} zodiac sign."

                return des
    
    class Perk(object):
        
        """This class is for all kinds of perks (used in the new perk system)"""
        
        def __init__(self, name, type, effects, archetype=None, pic = None, perk_level=0, min_rank=0, base_description = ""):
            self.name = name
            self.type = type
            self.effects = effects
            self.archetype = archetype
            self.level = perk_level
            self.min_rank = perk_level
            self.value = {0: 0, 1: 1, 2: 3, 3: 5}[perk_level]
            
            self.pic = pic
            if base_description:
                self.base_description = base_description
            else:
                self.base_description = perk_description[self.name]
            
        def get_pic(self):
            if self.pic:
                return Picture(self.pic, "perks/" + self.pic)
            else:
                return None
                
        def get_effect(self, type, target):
            for e in self.effects:
                if e.type.lower() == type.lower() and e.target.lower() == target.lower():
                    return True
            return False
            
        def get_description(self, short=False):
            if short:
                return get_description("", self.effects)
            else:
                return get_description("{i}" + self.base_description + "{/i}", self.effects)
    
    class PerkArchetype(object):
        
        """This class is for perk archetypes (used in the new perk system)"""
        
        def __init__(self, name, pic):
            
            self.name = name
            self.pic = pic
            self.unlocked = False
            self.base_description = archetype_description[self.name]
            
        def get_pic(self, portrait=False):
            if portrait:
                return Picture(self.pic[:-4] + " portrait" + self.pic[-4:], "perks/" + self.pic[:-4] + " portrait" + self.pic[-4:])
            else:
                return Picture(self.pic, "perks/" + self.pic)
            
        def get_perks(self, rank=None):
            if rank != None:
                return [perk for perk in perk_dict.values() if (perk.archetype == self.name and perk.level == rank)]
            else:
                mylist = [perk for perk in perk_dict.values() if (perk.archetype == self.name)]
                mylist.sort(key=lambda x: x.level)
                
                return mylist
            
        def get_description(self):
            return self.base_description
            
            
#        def init_perks(self):
#            self.slots = {}
            
#            for i in xrange(4):
#                self.slots[i] = archetype_perk_dict[self.name + " " + str(i)]
        
    
#        def has_effect(type, parameter): ## Syntax for effects is a tupple (type, parameter, value per level)
            
#            for eff in effects:
#                if eff(0) == type and eff(1) == parameter:
#                    return True
#            else:
#                return False
                    
#        def get_effect(type, parameter):
            
#            value = 0
            
#            for eff in effects:
#                if eff(0) == type and eff(1) == parameter:
#                    value += eff(2)*self.level

#            return value
            
                
    class Effect():
        
        """This class is used for all effects applying to a character"""
         
         ## Type defines how the effect work
         
         # Boost applies a % increase (or decrease). Value is a float number
         # Change applies a fixed value change which is not limited by stat max. Change can be reversed. Value is a number. 
         # Gain applies a one time permanent gain and is limited by stat max. Gain cannot be reversed. Value is a number.
         # Set replaces a base value with the new value
         # Allow unlocks a brothel option
         # Gift is the property of a gift item. I will have a different effect depending on a girl's tastes. Value is an int representing the gift bonus.
         # Flower is the property of a flower item. Value isn't used.
         # Special is hard-coded
         
         ## Value depends on the effect type. It is often used for checking the presence of an effect, so set it to 1 unless you need it to work differently
         
         ## Target defines what the effect affects
         
         ## Chance is the chance that the effect will happen. A float number.
         
         ## Scales_with is hard-coded for the moment and only concerns MC stats
         
         ## Scope is the scope of the effect: individual (None) or brothel-wide ("brothel")
         
        def __init__(self, type, target = None, value = 0, chance = 1.0, scales_with = None, scope = None, dice=False, change_cap=True):
            self.type = type
            self.target = target
            self.value = value
            self.chance = chance
            self.scales_with = scales_with
            self.scope = scope
            self.source = None
            self.dice = dice
            self.change_cap = change_cap
        
        def gain(self, thing, apply_boost=False, spillover=False):
        
            c = 0
            result = True

            #### APPLIES SCALING EFFECTS ####

            if self.scales_with:
                factor = get_scale_factor(thing, self.scales_with)
            else:
                factor = 1
            
            #### TEST EFFECT PROC CHANCE ####
            
            # Chance is tested here in case the gain is conditional
            if renpy.random.random() > self.chance:
                result=False
            
            #### APPLY EFFECT IF SUCCESSFUL ####
            
            # Instant cleaning (brothel)
            elif self.target == "dirt":
                c = brothel.change_dirt(factor*self.value)

            # Instant healing
            elif self.target == "heal":
                c = thing.heal(self.value)

            # Instant XP gain
            elif self.target in ("xp", "experience"):
                c = thing.change_xp(factor*self.value, apply_boost=apply_boost, spillover=spillover)

                if thing.ready_to_level():
                    thing.level_up()

            # Instant prestige gain
            elif self.target == "prestige":
                c = thing.change_prestige(factor*self.value, apply_boost=apply_boost)

                if thing.ready_to_level():
                    thing.level_up()

            # Instant JP gain                
            elif self.target[-3:] == " jp":
                c = thing.change_jp(factor*self.value, self.target[:-3], apply_boost=apply_boost, spillover=spillover, announcement_delay=0)
            
            # Instant REP gain
            elif self.target in ("rep", "reputation"):
                c = thing.change_rep(self.value)    
                                      
            # Instant perk points gain
            elif self.target in ("perk", "perks"):
                thing.perk_points += factor*self.value
                c = factor*self.value

            # Instant skill point gain
            elif self.target in ("skills", "skill points"):
                thing.upgrade_points += factor*self.value
                c = factor*self.value

            # Instant positive/negative fixation gain
            elif self.target.endswith(" fixation"):
                if self.value in fix_dict.keys():
                    c = thing.add_random_fixation(fixation=self.value, type=self.target[:3])
                else:
                    c = thing.add_random_fixation(type=self.target[:3])
            
            # Instant sex preference gain
            elif self.target == "all sexual preferences":
                for act in extended_sex_acts:
                    thing.change_preference(act, self.value)
            
            elif self.target.endswith(" preference"):
                thing.change_preference(self.target[:-11], self.value)
            
            # Permanent stat/skill gains
            else:
                c = thing.change_stat(self.target, factor*self.value, apply_boost=apply_boost, spillover=spillover)

            # Plays a sound when activating gain/instant
            if result:
                renpy.play(s_spell, "sound")
            else:
                renpy.play(s_fizzle, "sound")
            
            return c
        
        def get_description(self):
             
            value = self.value
            target = self.target
            text1 = ""

            if self.type in ("special", "personality"):
                
                if target == "level":
                    
                    text1 = "+1 level (max level: " + str(value) + ")"
                    
                elif target == "heal":
                    text1 = "accelerates a girl's healing by " + str(value) + " day(s)."
                
                elif target == "heal minion":
                    text1 = "Heals a wounded minion."
                
                elif target == "workwhore":
                    text1 = "she may work a half-shift, then whore a half-shift."
                
                elif target == "lucky":
                    text1 = "higher chance of critical success when working or whoring (does not stack)"
                    
                elif target == "unlucky":
                    text1 = "higher chance of critical failure when working or whoring"
                    
                elif target == "temptress":
                    text1 = "may talk an unwilling customer into accepting a different sex act"
                
                elif target == "pickpocket":
                    text1 = "chance of stealing an extra 10% tip from a customer, lowers reputation if caught"
                
                elif target == "random item":
                    text1 = "Customers have a small chance to 'forget' a random item in her care."
                    
                elif target == "BBCR bonus":
                    text1 = "may get a customer satisfaction boost if her beauty, body, charm or refinement skill is high enough"
                
                elif target == "LOCS bonus":
                    text1 = "may get a customer satisfaction boost if her libido, obedience, constitution or sensitivity skill is high enough"
                    
                elif target == "whore mood modifier":
                    text1 = "mood increase when whoring"
                    
                elif target == "job prestige":
                    text1 = "may earn prestige when working"
                    
                elif target == "skill catch up":
                    text1 += "every night, she will help other girls with lower skills receive a permanent skill increase (one girl per rank) "
                
                elif target == "effect chance":
                    text1 += "doubles the chances of every perk and trait activating for this girl"
                    
                elif target == "defender":
                    text1 += "you can defend the brothel even when you are out of AP"
                    
                elif target == "snake eyes":
                    text1 += "hypnosis never fails"
                    
                elif target == "safe":
                    text1 += "Excludes up to " + str(value) + " gold from the brothel threat level."
                
                elif target == "focus":
                    text1 += "+25% to tip and reputation gains if she has only one activated sex act (not including bisexual and group)."
                    
                elif target == "rest shield":
                    text1 += "When resting, may cast a magic shield on herself or a friend to protect from attacks." 
                
                return text1
            
            if self.type == "set":
                text1 += "change " + target + " to " + str(value)
                return text1
            
            if self.type == "allow":
                if target.endswith("preference"):
                    text1 += "Allows you to increase customers' " + target + " by up to +" + str(50*value) + "%."
                else:
                    text1 += "Allows " + target + " to visit your brothel."
                
                return text1
            
            if 0.75 <= self.chance < 1.0:
                text1 += "high chance of "
                
            elif 0.25 < self.chance < 0.75:
                text1 += "chance of "
            elif self.chance <= 0.25:
                text1 += "low chance of "
            
            
            if self.type == "reroll":
                if text1:
                    text1 += "rerolling "
                else:
                    text1 += "reroll"
                
                if self.target == "job critical failure":
                    text1 += " a critical failure when working"
                    
                return text1
            
            if self.dice:
                text1 += "1-"
            
            elif value > 0:
                text1 += "+"
            
            if self.type in ("gain", "instant"): # Permanent x gain (xp, reputation...)
                try:
                    text1 += str(round_int(value)) + " "
                except:
                    text1 += str(value) + " "
                    
                if self.target.endswith("preference") or self.target.endswith("preferences"):
                    text1 += " to "
            
            elif self.type == "change": # Temporary x effect (can be removed)
                text1 += str(round_int(value)) + " to "
            
            elif self.type == "resist":
                text1 += str(round_int(value)) + " negated "
            
            elif self.type == "spillover":
                percentage = round_int(value * 100)
                    
                text1 += str(percentage) + "% " + self.target + " spread out between other girls when earning "
            
            elif self.type == "boost": # Temporary % effect (can be removed)
                
                percentage = round_int(value * 100)
                    
                text1 += str(percentage) + "% to "
            
            elif self.type == "gift":
                text1 += str(round_int(value)) + " "
                
            elif self.type == "increase satisfaction":
                text1 += str(round_int(value)) + " to customer satisfaction for "
                
            if self.scope:
                text1 += self.scope + " "
            
            text1 += target
            
            if target == "hurt":
                text1 += " damage"
            elif target in extended_sex_acts:
                text1 += " acts"
            elif target == "random item":
                text1 += " when working"
            
            if self.scales_with:
                
                if self.scales_with == "equipped":
                    text1 += " for each equipped item"
                    
                elif self.scales_with == "cust nb":
                    text1 += " for each customer"
                elif self.scales_with == "job cust nb":
                    text1 += " for each customer when working"
                elif self.scales_with == "whore cust nb":
                    text1 += " for each customer when whoring"
  
                else:
                    text1 += " for each point of " + self.scales_with
            
            return text1

        
    class Sexact(): #Attributes: name, description, contributing_stats, variants, results
        
        """This class is for specifying sex acts and the results they have."""
        

        def __init__(self, likelihood, bonus):
            
            self.likelihood = likelihood
            self.bonus = bonus
        
    
    class Picture(object):
        
        """This class is for managing pictures and tags."""
        
        def __init__(self, filename="", path=""):
            
            if filename:
                self.filename = filename
            elif path:
                file_parts = path.split("/")
            
                self.filename = file_parts[-1]
            
            if not path:
                raise AssertionError, "No path provided for the Picture() object"
                
            self.path = path

            #<Chris12 PackState>
            self.make_tags_from_filename()
            self.__hashcode = None
            self.__filesize = -1
            #</Chris12 PackState>

            if self.filename.endswith(".webm"):
                self.video = True
            else:
                self.video = False
                    
        def get(self, x = None, y = None, proportional = True, side = False): # Doesn't work with just x or y for now
            
            if self.video:
                return Movie(play=self.path, channel="video")
            
            if side:
                return ProportionalScale(self.path, 150, 150, yalign = 1.0)
            
            if x and y:
                if proportional:
                    return ProportionalScale(self.path, x, y)
                else:
                    return im.Scale(self.path, x, y)
            else:
                return self.path
                
        
        def has_tag(self, tag): # Single tag search
            return self.has_tags([tag])
        
        def has_tags(self, tags, and_tags = [], not_tags = [], old = False): # Multiple tag (OR) search. For performance reasons, has_tags should always receive tuples as arguments
            
#            tags = make_tuple(tags)
#            and_tags = make_tuple(and_tags)
#            not_tags = make_tuple(not_tags)
            
            if tags:
                for tag in tags:
                    
                    if not tag: # Safety check in case a None value makes it into the tag list
                        continue
                    
                    elif (nsfw == False and tag in nsfw_tags) or (tag in (persistent.forbidden_tags + forbidden_tags)):
                        pass
                    
                    elif tag in self.tags or (old and tag in self.oldtags):
                        if not_tags:
                            for not_tag in not_tags:
                                if not not_tag:  # Safety check in case a None value makes it into the tag list
                                    continue
                                if not_tag in self.tags or (old and not_tag in self.oldtags):
                                    return False
                        
                        if and_tags:
                            for and_tag in and_tags:
                                if not and_tag:  # Safety check in case a None value makes it into the tag list
                                    continue
                                if (nsfw == False and and_tag in nsfw_tags) or (and_tag in (persistent.forbidden_tags + forbidden_tags)):
                                    pass
                                elif not (and_tag in self.tags or (old and and_tag in self.oldtags)):
                                    return False
                            else:
#                                renpy.say("", "Found " + self.filename + " with tags among " + and_text(tags) + " and " + and_text(and_tags) + " but not " + and_text(not_tags))
                                return True
                        else:
#                            renpy.say("", "Found " + self.filename + " with tags among " + and_text(tags))
                            return True
                        
            return False

        #<Chris12 PackState>

        # moved tagging to function, because it needs to be called from
        # __init__ as well as GirlFilesDict (during packstate rename)
        def make_tags_from_filename(self):
            self.tags = []
            self.oldtags = []
            lowerFilename = self.filename.lower()
            self.is_trash = ("_trash" in lowerFilename)
            self.is_unrecognized = ("_unrecognized" in lowerFilename)
            for tag in tag_dict: 
                if tag.lower() in lowerFilename:
                    self.oldtags.append(tag)
                    self.tags += make_tuple(tag_dict[tag])

        # calculates a checksum. 
        # Has to be the same as the function in the picture namer, so that packstates work
        def get_hash(self):
            if self.__hashcode is None :
                BLOCKSIZE = 65536
                hasher = hashlib.sha256()
                with open(config.gamedir + "/" + self.path, 'rb') as afile:
                    buf = afile.read(BLOCKSIZE)
                    while len(buf) > 0:
                        hasher.update(buf)
                        buf = afile.read(BLOCKSIZE)
                self.__hashcode = hasher.hexdigest()
            return self.__hashcode

        # gets the filesize
        # lazily initialized because its only needed by packstate check, not during normal gameplay
        def get_filesize(self):
            if self.__filesize < 0:
                self.__filesize = os.path.getsize(config.gamedir + "/" + self.path)
            return self.__filesize

        # Based on Goldo's Picture Namer
        def get_new_name(self):
            # self.tags.sort(key=sort_tags)
            
            if self.filename[-5:] in (".jpeg", ".webp"):
                ending = self.filename[-5:]
            else:
                ending = self.filename[-4:]

            if self.is_trash:
                new_file_name = "_TRASH " + " ".join(self.oldtags) + " (%s)" + ending
            elif self.is_unrecognized:
                tag_as_unrecognized = preferences.packstate_unrecognized != "Ignore"
                if self.filename.startswith("_UNRECOGNIZED ") :
                    if tag_as_unrecognized : 
                        new_file_name = self.filename # already has the tag
                    else :
                        new_file_name = self.filename[len("_UNRECOGNIZED "):] # remove the tag again
                else :
                    if tag_as_unrecognized :
                        new_file_name = "_UNRECOGNIZED " + self.filename
                    else :
                        new_file_name = self.filename # don't change the name if ignore
            elif self.oldtags:
                new_file_name = " ".join(self.oldtags) + " (%s)" + ending
            else:
                new_file_name = "_UNTAGGED" + " (%s)" + ending

            # When using old filename, put the (%s) into the new_file_name
            if "(" in new_file_name :
                new_file_name = new_file_name[:new_file_name.rfind("(")] + "(%s)" + ending
            else :
                new_file_name = new_file_name[:-len(ending)] + " (%s)" + ending
            return new_file_name

        # Based on Goldo's Picture Namer
        def commit_changes(self):
            
            i = 0
            new_file_name = self.get_new_name()
            
            # Compares file names up to the first parenthesis
            
            if new_file_name[:new_file_name.find("(")] != self.filename[:self.filename.find("(")]:
                
                while True :
                    if self.rename_file(new_file_name % str(i).zfill(5)):
                        return 1
                    else:
                        i += 1
                
            return 0

        # Based on Goldo's Picture Namer
        def rename_file(self, name):
            girlpath = self.path[:self.path.rfind("/")]
            os.chdir(config.gamedir + "/" + girlpath)
            
            try:
                os.rename(self.filename, name)
            except:
                return False
            
            self.filename = name
            self.path = girlpath + "/" + name
            
            return True

        #</Chris12 PackState>


    class ItemType(object):
        
        """ This class covers common item types and their base properties """
        
        def __init__(self, name, usage = "wear", slot = None, filter = "misc", sound = None, adjectives = "misc", stackable = False, dir=None):
            
            self.name = name
            self.usage = usage
            self.slot = slot
            self.filter = filter
            self.sound = sound
            self.adjectives = adjectives
            if dir:
                self.dir = dir
            else:
                self.dir = self.name
#            self.stackable = stackable
        
        
    class Item(object):
        
        """This class is for inanimate objects that the MC or girls can own."""
     
        def __init__(self, name, target, type, pic = None, template = False, rank = 1, max_rank = 5, rarity = 1, charges = None, price = 10000, effects = None, description = "", adjectives = None, sound = None, hidden_effect = False):
            
            self.base_name = name
            self.name = name
            self.target = target
            self.type = type
            
            if pic:
                self.pic = Picture(pic, "items/" + self.type.dir + "/" + pic)
            else:
                self.pic = Picture("misc.webp", "items/misc/misc.webp")
            
            self.template = template
            self.min_rank = rank
            self.rank = rank
            self.max_rank = max_rank
            self.rarity = rarity
            self.charges = charges
            self.base_price = price
            self.price = price
            self.base_effects = effects
            self.effects = effects
            self.equipped = False
            self.hidden_effect = hidden_effect
            
            ## Inherits properties from item type
            
            self.usage = self.type.usage
            self.slot = self.type.slot
            self.filter = self.type.filter
#            self.stackable = self.type.stackable
            
            ## An individual item can override type adjectives and sound if necessary
            
            if adjectives:
                self.adjectives = adjectives
                
            else:
                self.adjectives = self.type.adjectives
            
            if sound:
                self.sound = sound
                
            else:
                self.sound = self.type.sound
                
            self.base_description = description
            self.update_description()
        
        def get_key(self):
            return (self.type.name, self.rank, self.base_name, self.price)
        
        def has_effect(self, type="any", target="any"):
            for eff in self.effects:
                if (type in (eff.type, "any")) and (target in (eff.target, "any")):
                    return True
            return False
        
        def get_effect(self, type, target):
            return get_effect(self, type, target, iterate=True)
            
#            for eff in self.effects:
#                if type.lower() == eff.type.lower() and target.lower() == eff.target.lower():
#                    return eff.value
#            return False
        
        def can_wear(self, type):
            
            if self.usage != "wear":
                
                return False
                
            elif self.target == type:
                return True
                
            else:
                return False
        
        
        def can_use(self, type):
            
            if self.usage not in ("use", "auto", "buff"):
                
                return False
                
            elif self.target == type:
                return True
                
            else:
                return False
                
                
        def use_me(self, nb = 1):
            
#            renpy.say("", self.name +  " has " + str(self.charges) + " charges")
            
            if self.charges >= nb:
            
                self.charges -= nb
                
                if self.charges <= 0:
                    return "used_up"
                    
                else:
                    return self.charges
                    
            else:
                
                renpy.say("", "Not enough charges (" + str(self.charges) + ")")
                
                return "no charges"
            
            self.update_description()
        
        def get_price(self, operation):
            
            modifier = MC.get_modifier(operation)

            baseprice = self.price

            finalprice = round_int(baseprice * modifier)

            return finalprice
            
            
        def available_at_rank(self, rank): # Useless?
            
            if rank >= self.min_rank and rank <= self.max_rank:
                return True
            else:
                return False
        
        
        def transform(self, target_rank): #Transforms an item in a better or worse version of itself
            
            if self.template == True:
                self.name = quality_prefix[self.adjectives + "_" + str(target_rank)] + self.base_name.lower()
                self.price = round_int(quality_modifier[target_rank] * self.base_price)
                
                self.effects = []
                
                for eff in self.base_effects:
                    if target_rank > 0:
                        value = target_rank * eff.value
                    else:
                        value = eff.value / 2
                
                    self.effects.append(Effect(eff.type, eff.target, value))
                
                self.update_description()
                
                return self
                
            else:
                return None
        
        def generate_new_item(self, target_rank): # Creates new Item from this template item
            
            if self.template == True:
        
                new_it = copy.deepcopy(self)
                
                new_it.name = quality_prefix[self.adjectives + "_" + str(target_rank)] + self.base_name.lower()
                new_it.price = round_int(quality_modifier[target_rank] * self.base_price)
                
                if self.rarity in ("S", "U", "M"):
                    new_it.rarity = self.rarity
                else:
                    new_it.rarity = self.rarity + target_rank - self.min_rank
                
                new_it.effects = []
                
                for eff in self.effects:
                
                    if target_rank > 0:
                        value = target_rank * eff.value
                    else:
                        value = eff.value / 2
                
                    new_it.effects.append(Effect(eff.type, eff.target, value))
                
                new_it.min_rank = target_rank - 2
                new_it.rank = target_rank
                
                new_it.update_description()
                
#                renpy.say(new_it.name, "Min rank:" + str(new_it.min_rank) + " Rank:" + str(new_it.rank) + " Rarity:" + str(new_it.rarity))
                
                return new_it
        
        
        def update_description(self):
            
            if self.hidden_effect:
                
                self.description = self.base_description
                
            else:
            
                self.description = get_description(self.base_description, self.effects)
                
                if self.usage in ("use", "auto"):
                    
                    if self.charges > 1:
                        self.description += " (" + str(self.charges) + " uses left.)"
                
                elif self.usage == "buff":
                    
                    self.description += " (Lasts for "
                    
                    if self.charges >= 1:
                        self.description += str(self.charges) + " turns.)"
                        
                    elif self.charges == 1:
                        
                        self.description += "1 turn.)"
                
            
    
#    class Inventory(): # Attributes: Type (girl or items), items, girls, money
        
#        """This class is for inventories both with items and girls."""
        
    

    
    class Room(object):
        
        """This class is for rooms in a brothel, normal (bedrooms) and special (commons)"""
        
        def __init__(self, name, level = 0, type = "bedroom", job = None, cost=0): # Level is used for price of special rooms
            
            self.name = name
            self.level = level # Should be replaced with rank
            self.type = type # bedroom or special
            self.cust_limit = 0
            self.pic_path = "brothels/rooms/" + room_pics[self.name]
            self.pic = Picture(path=self.pic_path)
            self.bg = renpy.image(self.name, im.Scale(self.pic_path, config.screen_width, config.screen_height))
            self.job = job
            self.cost = cost
            self.girls = [] # Used for the master bedroom
            
        def add_girl(self, girl): # Used for the master bedroom
            if len(self.girls) < self.level:
                self.girls.append(girl)
                return True
            else:
                return False
            
        def remove_girl(self, girl): # Used for the master bedroom
            if girl in self.girls:
                self.girls.remove(girl)
                return True
            return False
            
        def can_have_girl(self, number=1):
            if len(self.girls) + number <= self.level:
                return True
            return False
        
        def buy(self, forced=False):
            if forced:
                self.level = 1
                self.update_cust_limit(True) 
            
            elif brothel.free_room:
                if renpy.call_screen("yes_no", "Do you really want to choose the " + self.name + " as your free room?"):
                    renpy.play(s_spell, "sound")
                    self.level = 1
                    self.update_cust_limit()
                    brothel.free_room = False
            
            elif self.get_price() >= MC.gold:
                renpy.say(sill, "Sorry Master, you do not have enough gold to build this room.")
            
            
            elif renpy.call_screen("yes_no", "Are you sure you want to build the " + self.name + " for " + str(self.get_price()) + " gold?"):
                MC.gold -= self.get_price()
                renpy.play(s_gold, "sound")
                self.level = 1
                self.update_cust_limit()
                
            unlock_pic(self.pic_path)
        
        def upgrade(self, forced = False):
            if forced:
                self.level += 1
                self.update_cust_limit(True)
            elif self.get_price() >= MC.gold:
                renpy.say(sill, "Sorry Master, you do not have enough gold to upgrade this room.")
            elif renpy.call_screen("yes_no", "Are you sure you want to upgrade the " + self.name + " for " + str(self.get_price()) + " gold?"):
                MC.gold -= self.get_price()
                renpy.play(s_gold, "sound")
                self.level += 1
                self.update_cust_limit()
                
        def get_pic(self, x, y, proportional=True):
            return self.pic.get(x, y, proportional)
            
        def get_price(self):
            if self.cost:
                return self.cost
            elif brothel.free_room:
                return 0
            else:
                return (self.level+1) * (150 + 150*game.chapter)
        
        def get_description(self):
            if self.type == "master":
                if self.level > 1:
                    desc = "Can host up to " + str(self.level) + " girls for training. Girls in the master bedroom receive free training every night."
                elif self.level == 1:
                    desc = "Can host up to 1 girl for training. Girls in the master bedroom receive free training every night."
                else:
                    desc = "Your bachelor pad. No room for a bachelorette yet."
                
                if self.level > 0:
                    desc += "\nCurrently assigned: "
                    if len(self.girls) >= 1:
                        desc += and_text([g.fullname for g in self.girls])
                    else:
                        desc += "No girl assigned"
                
                if self.level < brothel.rank:
                    desc += "\n{i}Improve this room for " + str(master_bedrooms[self.level+1].cost) + " gold."
                elif self.level == 5:
                    desc += "\n{i}You cannot improve the master bedroom further."
                else:
                    desc += "\n{i}You need a new brothel license to improve this room further."
                
                return desc
                
            if brothel_firstvisit:
                return "Build the " + self.name + " to train a " + self.job
            elif self.level == 0:
                return "Build the " + self.name + " for " + str(self.get_price()) + " gold."
            else:
                return "The " + self.name + " can host " + str(self.cust_limit) + " customers every night."
            
        def update_cust_limit(self, silent=False):
            self.cust_limit = room_capacity_dict[game.chapter] * self.level
                
            #<Chris Job Mod>
            if game.has_active_mod("chrisjobmod"):
                self.cust_limit = round_int(self.cust_limit * act_max_customers_modifier[self.job])
            #</Chris Job Mod>
            
            if not silent:
                renpy.say(sill, "You may now entertain " + str(self.cust_limit) + " customers in the " + self.name + ". {w=1.0}{nw}")

            
            
    class Quest(object):
        
        """This class is a template for quests and classes"""
        
        def __init__(self, type, name, main_stat, second_stat, other_stats, tags, description, sound = s_sigh, special_event = (None, 1.0)):
            
            self.type = type
            self.name = name
            self.description = description
            self.tags = make_tuple(tags)
            self.sound = sound
            
            self.main_stat = main_stat
            self.secondary_stat = second_stat
            self.other_stats = other_stats
            
            self.enrolled = []
            self.special = None
            self.special_event = special_event[0] # To be implemented later
            self.special_event_chance = special_event[1] # To be implemented later
            
        def randomize(self, rank):
            
            self.enrolled = []
            self.rank = rank
            
            ## Randomize picture
            
            self.pic = get_pic(quest_board, self.tags)
            
            
            ## Set duration
            
            self.duration = dice(rank + 2)
            
            
            ## Test special status
            
            if dice(6) == 6:
                if self.type == "class":
                    self.special = rand_choice(("Cheap", "Masterclass"))
                elif self.type == "quest":
                    self.special = rand_choice(("High reward", "Notorious"))
            else:
                self.special = None
            
            ## Set bonus and stat cap for classes
            
            if self.type == "class":
            
                self.bonuses = []
                
                self.bonuses.append([self.main_stat, self.duration*2, self.duration*3]) # Stores min/max bonuses to stat per day
                    
                if rank > 1 and dice(100) < (rank-1) * 25: # Chance of adding second bonus at higher ranks
                
                    self.bonuses.append([self.secondary_stat, self.duration, (self.duration)*2]) # Stores min/max bonuses to stat per day
                
                if rank > 3 and dice(100) < (rank-3) * 25: # Chance of adding third bonus at higher ranks
                    
                    self.bonuses.append([rand_choice(self.other_stats), self.duration-1, (self.duration-1)*3]) # Stores min/max bonuses to stat per day
                    
                self.stat_cap = 50 * rank
                
                self.capacity = dice(rank+2)
                
                
            ## Randomize stat requirements for quests
            
            if self.type == "quest":
            
                self.requirements = []
                
                # Normal stats have higher requirements than sx stats at earlier ranks then converge at rank 5
                if self.main_stat.capitalize() in gstats_main:
                    self.requirements.append([self.main_stat, 20 * rank + dice(50, rank)])
                    
                else:
                    self.requirements.append([self.main_stat, 4 * rank**2 + dice(50, rank)])
                    
                if dice(100) < (rank - 1) * 25: # Chance of adding second requirement at higher ranks
                        
                    if self.main_stat.capitalize() in gstats_main:
                        self.requirements.append([self.secondary_stat, 20 * (rank-1) + dice(50, rank-1)])
                        
                    else:
                        self.requirements.append([self.secondary_stat, 4 * (rank-1)**2 + dice(50, rank-1)])
                            
                if dice(100) < (rank - 2) * 25: # Chance of adding third requirement at higher ranks
                        
                    if self.main_stat.capitalize() in gstats_main:
                        self.requirements.append([rand_choice(self.other_stats), 20 * (rank-1) + dice(50, rank-1)])
                        
                    else:
                        self.requirements.append([rand_choice(self.other_stats), 4 * (rank-1)**2 + dice(50, rank-1)])
            
            
            ## Add 2 positive + 1 negative traits

            self.pos_traits = rand_choice(gold_traits + pos_traits, 2)
            
            self.neg_trait = rand_choice(neg_traits)
            
            
            ## Calculate rewards and costs
            
            if self.type == "quest":
                
                self.gold = 25*rank**2
                self.xp = 0
                self.rep = 0
                
                # Values have yet to be play-tested to make sure nothing is broken
                
                for stat, value in self.requirements:
                    
                    if stat in gstats_main:
                        self.gold += value * quest_base_gold["normal"]
                        self.rep += 2 ** (rank-1)
                        
                    else:
                        self.gold += value * quest_base_gold["sex"]
                        self.rep += 2 * (2 ** (rank-1))
                    
                    self.xp += value * rank
                    
                # Apply duration bonus (long quests bring more cash, short quests are good for rep)
                
                self.gold *= self.duration + 0.05*(self.duration-1)
                self.rep *= 1 + 0.05*(self.duration-1)
                self.xp *= self.duration
                
                if self.special == "High reward":
                    self.gold *= 1.5
                    
                self.gold = round_int(self.gold * brothel.get_effect("boost", "quest rewards"))
                self.rep = round_int(self.rep * brothel.get_effect("boost", "quest rewards"))
                self.xp = round_int(self.xp * brothel.get_effect("boost", "quest rewards"))
                
            elif self.type == "class":
                
                self.gold = 25*rank**2
                
                for stat, min, max in self.bonuses:
                    if stat in gstats_main:
                        self.gold += (min+max)/2 * 5
                    else:
                        self.gold += (min+max)/2 * 10
                
                self.xp = 25 * rank * self.duration
                self.rep = (1 + 0.05*(self.duration-1)) * 2**(rank-1)
                
                if self.special == "Cheap":
                    self.gold = round_int(0.75*self.gold)
                    
            self.energy = -10 * rank * self.duration
                
                
        def test_eligibility(self, girl, ignore_status = False): # Returns a tuple: bool + ttip
            
            if not ignore_status:
                if girl.hurt > 0 or girl.away or girl.exhausted:
                    return (False, "Your girl is unable to work or study at the moment.")
            
            if self.type == "class":
                
                if MC.gold >= self.gold:
                
                    for stat, min, max in self.bonuses:
                        if girl.get_stat(stat, raw=True) < self.stat_cap:
                            return (True, "Register this girl for the selected class.")
                    
                    return (False, "Your girl's skills are too high to learn anything from this class.")
                    
                else:
                    return (False, "You do not have enough money to register a girl for this class.")

            
            elif self.type == "quest":
                
                for stat, value in self.requirements:
                    if girl.get_stat(stat) < value:
                        return (False, "Your girl doesn't meet the requirements for this assignment.")
                return (True, "Send this girl on an assignment.")
            
            raise AssertionError, "Something is weird with " + self.type
            
        def count_eligible_girls(self):
            
            eligible = sum(1 for girl in MC.girls if self.test_eligibility(girl, ignore_status=True)[0])
                    
            return eligible
            
        def get_results(self, girl):
            
            title = self.type.capitalize() + " completed"
            description = girl.fullname + " has returned from her " + self.type + ". "
            
            if self.type == "class":
            
                # Only mood affects learning for now
                perf = dice(6, 2) + girl.mood // 20
                
                # Memories of rewards and punishment
                
                if girl.remembers("reward", "class good result"):
                    perf += 1
                if girl.remembers("punish", "class bad result"):
                    perf += 1
                
                if perf >= 12:
                    description += "She studied very hard and made exceptional progress."
                    boost = 2.0
                    
                    girl.track_event("class good result", arg="She studied really hard for her " + self.name  + " class.")    
                    
                elif perf >= 9:
                    description += "She listened carefully to her teacher and made good progress."
                    boost = 1.5
                    
                    girl.track_event("class good result", arg="She made good progress during her " + self.name + " class.")   
                
                elif perf <= 2:
                    description += "She didn't care about the lessons at all, making almost no progress."
                    boost = 0.75
                    
                    girl.track_event("class bad result", arg="She didn't study hard during her " + self.name + " class.")
                    
                elif perf <= 5:
                    description += "She was barely paying attention to her teacher, hindering her progress."
                    boost = 0.5
                    
                    girl.track_event("class bad result", arg="She didn't study at all for her " + self.name + " class.")
                    
                else:
                    description += "She made some progress."
                    boost = 1.0
                
                if self.special == "Masterclass":
                    boost *= 1.5
                
                boost *= girl.get_effect("boost", "class results")
                
                changes = [(stat, round_int(girl.change_stat(stat, renpy.random.randint(min, max)*boost))) for stat, min, max in self.bonuses]
            
            elif self.type == "quest":
                
                # Only main stat affects result. Random variation from +6 to -6 times rank
                perf = girl.get_stat(self.main_stat) + girl.mood // 5 + dice(13, self.rank) - 7*self.rank - self.requirements[0][1]
                
                # Memories of rewards and punishment
                
                if girl.remembers("reward", "quest good result"):
                    perf += 1
                if girl.remembers("punish", "quest bad result"):
                    perf += 1
                
                if perf >= 20 * self.rank:
                    description += "{color=[c_orange]}Her performance was amazing. The customer was ecstatic!{/color}"
                    boost = 1.5
                    
                    girl.track_event("quest good result", arg="She performed {color=[c_emerald]}amazingly well{/color} on a quest")
                    
                elif perf >= 10 * self.rank:
                    description += "{color=[c_emerald]}She performed well. The customer was happy.{/color}"
                    boost = 1.25
                    
                    girl.track_event("quest good result", arg="She performed {color=[c_emerald]}well{/color} on a quest")
                    
                elif perf < 0:
                    description += "{color=[c_red]}She performed poorly. The customer was disappointed and refused to pay in full.{/color}"
                    boost = 0.75
                    girl.track_event("quest bad result", arg="She performed {color=[c_crimson]}badly{/color} while on a quest")
                    
                else:
                    description += "She completed the assignment without incident."
                    boost = 1.0
                
                had = []
                
                for trait in self.pos_traits:
                    if girl.has_trait(trait.name):
                        had.append(trait.get_past_tense())
                        boost *= 1.5
                        
                if had:
                    description += " The customer was excited that she " + and_text(had) + "."
                
                if girl.has_trait(self.neg_trait.name):
                    description += " The customer was upset that she " + self.neg_trait.get_past_tense() + "."
                    boost /= 2.0
                
                boost *= girl.get_effect("boost", "quest results")
                
                reward = round_int(self.gold*boost)
                self.rep = round_int(self.rep*boost)
                
                if self.special == "Notorious":
                    self.rep *= 2.0
                
                MC.gold += reward
                girl.add_log("quest_gold", reward, -1)
                
            xp = round_int(girl.change_xp(self.xp))
            rep = round_int(girl.change_rep(self.rep))
            energy, status = girl.change_energy(self.energy)
                
            if status == "exhausted":
                " She is {color=[c_red]}exhausted{/color} and will need to rest until she recovers."
            
            description += "\n"
            
            girl.add_log("total_xp", xp, -1)
            
            if self.type == "class":
                girl.add_log("class_xp", xp, -1)
                
                for stat, value in changes:
                    description += stat_increase_dict["stat"] % (stat.capitalize(), value)
            
            elif self.type == "quest":
                girl.add_log("quest_xp", xp, -1)
                
                description += stat_increase_dict["gold+"] % reward
            
            description += stat_increase_dict["xp_dark"] % xp
            description += stat_increase_dict["rep"] % rep
            description += stat_increase_dict["stat_neg"] % ("Energy", energy)
            
            if girl.ready_to_level():
                girl.level_up()
                description += stat_increase_dict["level"]

            if girl.ready_to_rank():
                description += stat_increase_dict["rank"]
            
            return title, description
            


    class Goal(object):
        
        def __init__(self, type, value = 0, target = 0):
            
            self.type = type
            self.value = int(value)
            self.target = int(target)
            self.description = self.get_description()
        
        def get_description(self):
            
            if self.type == "gold":
                return "you must collect " + str(self.value) + " gold"
                
            elif self.type == "ranked":
                return str(self.target) + " of your girls must reach rank " + rank_name[self.value]
                
            elif self.type == "reputation":
                return "your brothel must reach " + str(self.value) + " reputation"
                
            elif self.type == "prestige":
                return "you must gather " + str(self.value) + " prestige"
                
            else:
                return ""
        
        def reached(self):
            
            if self.type == "gold":
                
                if MC.gold >= self.value:
       
                    return True
                
                else:
                    
                    return False
            
            elif self.type == "ranked":
                
                ranked_girls = 0
                
                for girl in MC.girls:
                    if girl.rank >= self.value:
                        ranked_girls += 1
                
                if ranked_girls >= self.target:
                    return True
                else:
                    return False
            
            elif self.type == "reputation":
                
                if brothel.rep >= self.value:
                    return True
                else:
                    return False
                    
            elif self.type == "prestige":
                
                if MC.prestige >= self.value:
                    return True
                else:
                    return False
                    
            else:
                return False
    
    
    
    class Location(object):
        
        """This class is used for locations you can visit in the various city districts"""
        
        def __init__(self, name, pic, has_girls = True, secret = False, action = False, menu = None, menu_costs_AP = True):
            
            self.name = name
            self.pic = "districts/locations/" + pic
            self.has_girls = has_girls
            self.secret = secret
            self.action = action
            self.menu = menu # Menu is a tuple: first is the button caption, second is the target label
            self.menu_costs_AP = menu_costs_AP
            self.girls = []
            self.runaways = []
#            self.bg = renpy.image("bg " + name.lower(), ProportionalScale(self.pic, config.screen_width, int(config.screen_height*0.8)))
            
        def get_pic(self, x, y):
        
            return im.Scale(self.pic, x, y)
        
        def list_girls(self):
            
            l = []
            
            for g in self.girls:
                
                l.append(g.name)
                
            return "This location has girls: " + and_text(l)
            
        def clear_girls(self):
            
            del self.girls[:]
            
            return
        
        def can_do_action(self):
            if not self.action:
                return False
            if self.menu_costs_AP and MC.interactions < 1:
                return False            
            if self.menu[1].startswith("collect_"):
                if MC.last_collected[self.menu[1][8:]] == calendar.time:
                    return False
            elif self.menu[1] == "visit_thieves_guild":
                if not NPC_renza.items:
                    return False
            elif self.menu[1] == "visit_watchtower":
                if not NPC_captain.items:
                    return False
            return True
        
    
    class Spell(object):
        
        """ This class covers spells used by the MC """
        
        def __init__(self, name, pic = "aura1.webp", type = "passive", level = 0, cost = 0, effects = [], duration = None, sound = "spell.ogg", description = "A basic spell."):
            
            self.name = name
            self.pic = Picture(pic, "spells/" + pic)
            self.type = type
            self.level = level
            self.cost = cost # Passive spells have a cost of zero
            if effects:
                self.effects = effects
            else:
                self.effects = []
#             self.instant = instant
            self.duration = duration
            self.sound = sound
            self.description = get_description(description, effects)
#            self.learnt = False
#            self.active = False
            self.auto = False
            
        def get_cost(self):
            return self.cost
            
#             cost = self.cost - MC.get_spirit()
            
#             if cost < 1:
#                 cost = 1
                
            
        def get_cost_description(self):
            if self.duration == "turn":
                return str(self.get_cost()) + " mana/turn"
                
            else:
                return str(self.get_cost()) + " mana"
        
        
    class Personality(object):
        
        def __init__(self, name, description, attributes, allies, rivals):
            
            self.name = name
            self.description = description
            self.attributes = attributes
            self.allies = allies
            self.rivals = rivals
            
        def has_attributes(self, attr):
            if attr == None:
                return True
            
            attr = make_tuple(attr)
            
            for a in attr:

                if a not in gpersonality_attributes:
                    raise AssertionError, "Attribute '" + a + "' not found"
                
                if not (a in self.attributes or "very " + a in self.attributes):
                    return False
            
            return True
    
    
    class Fixation(object):
        
        def __init__(self, name, acts, step, frequency = 12.0, tag_list = [], not_list = [], attribute = None, short_name=""):
            
            self.name = name
            if short_name:
                self.short_name = short_name
            else:
                self.short_name = self.name
            self.acts = make_tuple(acts)
            self.step = step
            self.frequency = frequency
            self.tag_list = tag_list
            self.not_list = not_list
            self.attribute = attribute
#            self.perk = sexual_perks[name]
        
        def available(self, girl, act=None):
            if self.name in girl.init_dict["sexual preferences/never_fixations"]:
                return False
            
            if act and act not in self.acts:
                return False
            
            if not girl.is_(self.attribute):
                return False
            
            if self in (girl.pos_fixations + girl.neg_fixations):
                return False
            
            return True
    
        def get_weight(self, girl, type="pos"):
            
            freq = self.frequency
            
            if self.attribute:
                if girl.is_(self.attribute):
                    freq *= 3
            
            if self.name in girl.init_dict["sexual preferences/favorite_fixations"]:
                freq *= 3
            elif self.name in girl.init_dict["sexual preferences/disliked_fixations"]:
                freq /= 3
            else:
                for act in self.acts:
                    if act in girl.init_dict["sexual preferences/favorite_acts"]:
                        freq *= 2
                    elif act in girl.init_dict["sexual preferences/disliked_acts"]:
                        freq /= 2
            
            if type == "neg":
                freq = 100/freq
            
#            renpy.say("", str(freq))
            
            return int(freq)
    
    
    class GirlInteractionTopic(object):
        
        def __init__(self, type, group, caption, label, AP_cost=1, gold_cost=0, act=None, condition=None, advanced=False, love_test=None, relationship_level=0):
            
            self.type = type # Determines which tab the option belongs to in the interaction menu
            self.group = group # Determines which group the interaction is counted a part of (e.g. all training shares the 'train' group)
            self.caption = caption
            self.label = label
            self.AP_cost = AP_cost
            self.gold_cost = gold_cost
            self.act = act # For training and magic training
            self.advanced = advanced # Determines if advanced options are available
            self.condition = condition
            self.love_test = love_test
            self.relationship_level = relationship_level
        
        def get_gold_cost(self):
            return self.gold_cost*(district.rank ** 2)
        
        def is_shown(self, girl): # The option won't display unless the condition is True
            if self.love_test != None:
                if girl.get_love() + MC.get_charisma() >= self.love_test: # a value of 0 for love_test will be used in the test. Give love_test a 'None' value to ignore the test.
                    return True
                else:
                    return False
            
            if self.relationship_level:
                if girl.MC_relationship_level >= self.relationship_level:
                    return True
                else:
                    return False
            
            if not self.condition:
                return True
            elif self.condition == "has_worked" and girl.has_worked:
                return True
            elif self.condition == "other_girls" and len(MC.girls) >= 2:
                return True
            elif self.condition == "story" and girl.flags["story"] == 50:
                return True
            elif self.condition == "neg_fix":
                neg_fix = [fix.name for fix in girl.neg_fixations if girl.personality_unlock[fix.name]]
                if neg_fix:
                    return True
            elif self.condition == "master_bedroom_add":
                if brothel.master_bedroom.level >= 1 and girl not in brothel.master_bedroom.girls:
                    return True
            elif self.condition == "master_bedroom_remove":
                if girl in brothel.master_bedroom.girls:
                    return True
            elif self.condition == "dressed":
                if girl.get_effect("special", "naked") and not girl.naked:
                    return True
            elif self.condition == "naked":
                if girl.naked:
                    return True
            elif self.condition == "farm":
                if farm.active:
                    return True
            elif self.condition == "debug_mode" and debug_mode:
                return True
            
            
            return False
                            
        def is_available(self, girl, mode=None, free=False): # The option will display inactive if False. Returns a tuple with bool and a tooltip description.
            
            if mode: # 'mode' is either 'lecture' (Talk), 'train' or advanced.  
                if mode == "lecture":
                    pass
                elif mode in ("train", "advanced"):
                    text1 = ""
                    
                    if self.type == "train":
                        if training_test_dict[self.act]:
                            for cond, pref in training_test_dict[self.act]:
                                if compare_preference(girl, cond, pref):
                                    break
                                if text1:
                                    text1 += " or "
                                text1 += cond + " (" + pref + ")"
                            else:
                                return False, "You cannot train " + self.act + " yet. Requirements: " + text1
                                
                    elif self.type == "magic":
                        if magic_training_test_dict[self.act]:
                            for cond, pref in magic_training_test_dict[self.act]:
                                if compare_preference(girl, cond, pref):
                                    break
                                if text1:
                                    text1 += " or "
                                text1 += cond + " (" + pref + ")"
                            else:
                                return False, "You cannot train " + self.act + " yet. Requirements: " + text1
                
                    if mode == "advanced":
                        if girl.personality_unlock[self.act]:
                            
                            pos_reaction, neg_reaction = girl.test_weakness(self.act)
                            
                            if not (pos_reaction or neg_reaction):
                                return False, "Advanced training is not available. She has no particular weakness to this sex act."
                            
                        else:
                            return False, "You need to train a girl at least once before you can access advanced training."
                            
                        if MC.interactions < 2 and not free:
                            return False, "You do not have enough interactions left for advanced training."
                            
                elif mode == "master_bedroom_add":
                    if not brothel.master_bedroom.can_have_girl():
                        return False, "The master bedroom is already full."
                                                   
            if MC.interactions < 1 and self.AP_cost > 0 and not free:
                return False, "You have no interactions left for today."
            elif MC.interactions < self.AP_cost and not free:
                return False, "You do not have enough interactions left for this."
            elif self.get_gold_cost() and MC.gold < self.get_gold_cost():
                return False, "You do not have enough money to pay for this training (" + str(self.get_gold_cost()) + "{image=img_gold})."
            
            elif self.group == "train" and girl.MC_interact_counters[self.group] >= 1:
                return False, "You cannot train a girl more than once per day."
            elif self.group in ("gold", "gift", "sex_reward", "rape", "offer") and girl.MC_interact_counters[self.group] >= 1:
                return False, "You cannot do that more than once per day."
                
#            elif self.label == "slave_reward_gold" and girl.MC_interact_counters["gold"] >= 1:
#                return False, "You cannot give her gold more than once per day."
#            elif self.label == "slave_reward_gift" and girl.MC_interact_counters["gift"] >= 1:
#                return False, "You cannot give her a gift more than once per day."
#            elif self.label == "slave_reward_sex" and girl.MC_interact_counters["sex"] >= 1:
#                return False, "You cannot give her a sexual reward more than once per day."
#            elif self.label == "slave_punish_rape" and girl.MC_interact_counters["rape"] >= 1:
#                return False, "You cannot rape her more than once per day."

            elif self.group == "offer" and len(MC.girls) >= brothel.bedrooms:
                return False, "You don't have room in your brothel for another girl."
            elif self.group and girl.MC_interact_counters[self.group] >= 3:
                return False, "You cannot " + self.group + " more than 3 times a day with a girl."
            elif self.label == "slave_master_bedroom_add" and not brothel.master_bedroom.can_have_girl():
                return False, "The master bedroom is already full."
            return True, ""
        
            
    
    class GirlInteraction(object):
        
        def __init__(self, girl, topic, mode=None, free=False):
            
            self.girl = girl
            self.topic = topic
            self.mode = mode
            self.free = free
            
            if self.free:
                self.cost = 0
            elif self.mode == "advanced":
                self.cost = 2
            else:
                self.cost = 1
            
            self.type = topic.type # This gets replaced by chat or react type in the course of the interaction (not stricly necessary, but easier to manipulate)
            self.act = topic.act
            
            self.response = None
            self.MC_reaction = None
            self.reason = None
            self.pic = None
            self.score = 0
            self.result = None
            self.other_girl = None
            self.canceled = False
            
        def resolve(self):
            
#            renpy.say("", "RESOLVING")
            
            if self.canceled:
                return

            girl = self.girl
            
            if debug_mode:
                renpy.call_screen("OK_screen", girl.fullname + "-Interaction resolving", "Group: [inter.topic.group]\nType: [inter.type]\nAction: [inter.topic.caption]\nReason: [inter.reason]\nResponse: [inter.response]\nAct: [inter.act]\nResult: [inter.result]\nMC Reaction: [inter.MC_reaction]")
            else:
                renpy.block_rollback()
            
                        
            # Charging interactions
            
            if self.result in ("moderate", "fail") or self.MC_reaction == "give up": # Failed advanced training attempts only cost 1 AP
                MC.interactions -= 1
            else:
                MC.interactions -= self.cost
            
            # Tracking interaction count
            girl.MC_interact_counters[self.topic.group] += 1
            
            if topic.group in ("gold", "gift", "sex_reward", "rape"):
                girl.MC_interact_counters["react"] += 1
            
            # Creating local variables for the stat changes
            m = 0
            l = 0
            f = 0
            p = 0
            bea = 0
            bod = 0
            cha = 0
            ref = 0
            ob = 0
            lib = 0
            sen = 0
            con = 0
            sv = 0
            sx = 0
            an = 0
            fe = 0
            gd = 0
            ev = 0
            ne = 0
            brk = defaultdict(int)
            virgin = False
            
            if self.topic.type == "chat":
                
                # In case she is too afraid to talk
                
                if self.response == "afraid":
                    if self.MC_reaction == "encourage":
                        f -= 1
                    elif self.MC_reaction == "discipline":
                        f += 1
                        
                else:
                    
                    ## 1. Mood change
                    
                    if self.MC_reaction != "discipline":
                        
                        # Extravert people like chatting
                        
                        if girl.is_("very extravert"):
                            m += 2
                        elif girl.is_("extravert"):
                            m += 1
                        elif girl.is_("very introvert"):
                            m -= 1
                        
                        # Mood improves when chatting (except if punished)
                        
                        m += 1
                        
                    else:
                        
                        # Impact is stronger on dom girls
                        
                        if girl.is_("very dom"):
                            m -= 2
                        elif girl.is_("dom"):
                            m -= 1
                        elif girl.is_("very sub"):
                            m += 1
                        
                        
                    
                    ## 2. Personality unlocking
                    
                    if self.type == "slave_life" or self.type == "feelings":
                        unlock = "DS"
                    elif self.type == "brothel" or self.type == "tastes":
                        unlock = "LR"
                    elif self.type == "customers" or self.type == "well_being":
                        unlock = "MI"
                    elif self.type == "other_girls" or self.type == "origins":
                        unlock = "EI"
                    else:
                        raise AssertionError, "Type not found. Type is " + str(self.type)
                    
                    girl.personality_unlock[unlock] += 15 + MC.get_charisma() + dice(10) # Temp, see how it behaves
                    
                    # Story unlocking, faster if chatting about personal stuff: First step unlocks at 4, then 10, then 20
                    
                    if self.type in ("well_being", "feelings", "tastes", "origins"):
                        girl.personality_unlock["story"] += 1 + 0.1 * MC.get_charisma()
                    else:
                        girl.personality_unlock["story"] += 0.5 + 0.05 * MC.get_charisma()
                    
                    
                    ## 3. Love and fear
                    
                    r = dice(6) + MC.get_charisma()
                    
                    if self.MC_reaction == "encourage": # Idealist girls like to be encouraged
                        
                        m += dice(3)
                        
                        if r >= 6:
                            f -= 1
                        
                        if r >= 9:
                            l += 1
                        
                        if girl.is_("very idealist"):
                            l += 0.5
                            f -= 1
                        elif girl.is_("idealist"):
                            l += 0.25
                            f -= 0.5
                        elif girl.is_("materialist"):
                            f -= 0.25
                        
                    elif self.MC_reaction == "discipline": # Dom girls dislike being disciplined
                        
                        m -= dice(3)
                        
                        if r < 6:
                            l -= 1
                        
                        if r >= 9:
                            f += 1
                        
                        if girl.is_("very sub"):
                            l += 0.5
                            f += 1
                        elif girl.is_("sub"):
                            l += 0.25
                            f += 0.5
                        elif girl.is_("very dom"):
                            l -= 0.5
                        elif girl.is_("dom"):
                            l -= 0.25
                        
                        
                    ## 4. Random skill increases
                    
                    r = dice(6) + MC.get_charisma()
                    
                    if self.type == "slave_life":
                        if r >= 9:
                            ob += 1
                            
                    elif self.type == "brothel":
                        if r >= 9:
                            lib += 1
                        
                    elif self.type == "customers":
                        if r >= 9:
                            sen += 1
                        
                    elif self.type == "other_girls":
                        if r >= 9:
                            cha += 1
                            
                    ## Impact other girls (if necessary)
                    
                    if self.other_girl:
                        
                        r = dice(6) + MC.get_charisma()
                        
                        if self.MC_reaction == "praise friend": # Extraverts and idealists like to be encouraged in their relationships
                            m += dice(3)

                            if r >= 6:
                                f -= 1
                            
                            if r >= 9:
                                l += 1
                            
                            if girl.is_("very extravert"):
                                l += 0.5
                                f -= 1
                            elif girl.is_("extravert"):
                                l += 0.25
                                f -= 0.5
                            elif girl.is_("introvert"):
                                f += 0.5
                                
                            if girl.is_("very idealist"):
                                l += 0.5
                            elif girl.is_("idealist"):
                                l += 0.25
                            elif girl.is_("materialist"):
                                l -= 0.5
                                
                            self.other_girl.change_love(1 * (1+MC.get_charisma()/10))
                            
                        elif self.MC_reaction == "demean rival": # Extraverts and materialists like to be encouraged in their feuds
                        
                            m += dice(3)

                            if r >= 6:
                                f -= 1
                            
                            if r >= 9:
                                l += 1
                            
                            if girl.is_("very extravert"):
                                l += 0.5
                            elif girl.is_("extravert"):
                                l += 0.25
                            elif girl.is_("introvert"):
                                f += 0.5
                                
                            if girl.is_("very materialist"):
                                l += 0.5
                            elif girl.is_("materialist"):
                                l += 0.25
                            elif girl.is_("idealist"):
                                l -= 0.5
                                
                            self.other_girl.change_love(-1 * (1+MC.get_charisma()/10))
                        
                        elif self.MC_reaction == "break friendship": # Extraverts and idealists hate to be discouraged in their relationships
                            
                            m -= dice(6)
                            
                            if r >= 6:
                                l -= 1
                            
                            if r >= 9:
                                f += 1
                            
                            if girl.is_("very extravert"):
                                l -= 2
                                f += 1
                            elif girl.is_("extravert"):
                                l -= 1
                                f += 0.5
                            elif girl.is_("introvert"):
                                l -= 0.5
                            
                            if girl.is_("very idealist"):
                                l -= 2
                                f += 1
                            elif girl.is_("idealist"):
                                l -= 1
                                f += 0.5
                            elif girl.is_("materialist"):
                                f += 0.5
                            
                            girl.change_relationship(self.other_girl, -1 * MC.get_charisma()/3)
                        
                            self.other_girl.change_love(-1)
            
                        elif self.MC_reaction == "make peace": # Extraverts and materialists don't like to be discouraged in their feuds
                            
                            m -= dice(3)
                            
                            if r >= 6:
                                l -= 1
                            
                            if r >= 9:
                                f += 1
                                
                            if girl.is_("very extravert"):
                                l -= 1
                                f += 0.5
                            elif girl.is_("extravert"):
                                l -= 0.5
                            elif girl.is_("introvert"):
                                f -= 0.5
                                
                            if girl.is_("very materialist"):
                                l -= 1
                                f += 0.5
                            elif girl.is_("materialist"):
                                l -= 0.5
                            elif girl.is_("idealist"):
                                l += 0.5
                            
                            girl.change_relationship(self.other_girl, MC.get_charisma()/3)
                            
                            self.other_girl.change_love(1)
                            
            
            elif self.topic.type == "train" and self.mode != "lecture": # Non-magical training, excluding lectures
                
                ## In case she refused training
                
                r = dice(6) + MC.get_charisma()
                                
                # MC gave up
                
                if self.MC_reaction == "give up": 
                    f -= 1
                    
                    if self.response == "begged":
                        gd += 1
                        l += 1
                        f -= 1
                            
                    elif self.response == "resisted":
                        ev -= 1 # MC gets less evil when not forcing
                                                
                        if r <= 6: # Girls may respect the player less
                            ob -= dice(3)
                    
                    if self.response != "refused": # Repressed girls are more relieved when MC gives up
                        if girl.is_("very repressed"):
                            l += 1
                            f -= 1
                        elif girl.is_("repressed"):
                            l += 0.5
                            f += 0.5
                        else:
                            ob -= 1
                                            
                elif self.MC_reaction == "warning": # Warnings work better on sub girls
                    ev -= 1
                    f += 1
                        
                    if self.response == "begged":
                        ev -= 1
                    
                    if girl.is_("very sub"):
                        f += 1
                        ob += 2
                    elif girl.is_("sub"):
                        f += 0.5
                        ob += 1
                    elif girl.is_("very dom"):
                        f -= 1
                        ob -= 2
                    elif girl.is_("dom"):
                        f -= 0.5
                        ob -= 1
                
                
                # MC got into a fight and lost
                
                elif self.response == "refused" and self.result == "fail":
                    ev += 1
                    l -= 5
                    f -= 5
                
                ## If she went on with the training
                
                else:
                    ## 1. Mood change
                    
                    m -= 1
                    
                    if girl.fear >= girl.love + 10: # She doesn't like interacting with MC
                        m -= 1
                    elif girl.love >= girl.fear + 10: # She likes the MC
                        m += 1
                    
                    if self.MC_reaction == "encourage":
                        m += 1
                    elif self.MC_reaction == "discipline":
                        m -= 1
                    
                    if self.response == "begged": # If MC ignored her begging
                        m -= 3
                    
                    ## 2. Stat changes (normal training)
                    
                    if self.act in ("obedience", "constitution", "lecture"):
                        
                        r = dice(6) + MC.get_charisma()
                        
                        if self.act == "obedience":
                            if self.result == "good":
                                ob += dice(6)
                            elif self.MC_reaction == "discipline": # Scolding her might have a good effect
                                if r >= 9:
                                    ob += dice(3)
                            
                        elif self.act == "constitution":
                            if self.result == "good":
                                con += dice(6)
                            elif self.MC_reaction == "discipline": # Scolding her might have a good effect
                                if r >= 9:
                                    con += dice(3)
                    
                    # 3. Love and fear (sexual training)
                    
                    else:
                        
                        r = dice(6) + MC.get_charisma()
                        
                        if self.response == "accepted": # Lewd girls like consensual sex, repressed girls are reassured
                            
                            f -= 0.5
                            
                            if girl.is_("very lewd"):
                                l += 1
                            elif girl.is_("lewd"):
                                l += 0.5
                            elif girl.is_("very repressed"):
                                f -= 1
                            elif girl.is_("repressed"):
                                f -= 0.5
                        
                        elif self.response == "resisted": # Dom and repressed girls hate being forced, sub and lewd girls don't mind too much
                            ne += 1
                            
                            f += 1
                            
                            if girl.is_("very dom"):
                                l -= 1
                            elif girl.is_("dom"):
                                l -= 0.5
                            elif girl.is_("very sub"):
                                l += 1
                            elif girl.is_("sub"):
                                l += 0.5
                            
                            if girl.is_("very repressed"):
                                f += 1
                            elif girl.is_("repressed"):
                                f += 0.5
                            elif girl.is_("very lewd"):
                                f -= 1
                            elif girl.is_("lewd"):
                                f -= 0.5
                        
                        elif self.response == "begged": # If MC ignored her begging
                            gd -= 1
                            f += 1
                            
                            if girl.is_("very sub"): # Reminder: Only sub girls may beg
                                l += 0.5
                                f += 1
                            elif girl.is_("sub"):
                                f += 0.5
                                                            
                            if girl.is_("very repressed"):
                                f += 1
                            elif girl.is_("repressed"):
                                f += 0.5
                            elif girl.is_("very lewd"):
                                f -= 1
                            elif girl.is_("lewd"):
                                f -= 0.5
                        
                        else: # (forced) Dom and repressed girls hate being forced. Very sub and very lewd girls kind of like it
                            ev += 1
                            
                            f += 3
                            
                            if girl.is_("very dom"):
                                l -= 2
                            elif girl.is_("dom"):
                                l -= 1
                            elif girl.is_("very sub"):
                                l += 0.5
                                
                            if girl.is_("very repressed"):
                                f += 2
                            elif girl.is_("repressed"):
                                f += 1
                            elif girl.is_("very lewd"):
                                f -= 1
                        
                        # Obedience may increase during s. training
                        
                        if r >= 9:
                            ob += dice(3)
                        elif r >= 6:
                            ob += 1

            elif self.topic.type == "magic": # Magical training
                
                if self.result == "success": # Note: stat changes and breaking handled with sex act results below
                    
                    if girl.magic_training == "positive":
                        l += 1
                    elif girl.magic_training == "negative":
                        f += 1
                    
                    if self.act == "obedience":
                        ob += dice(6)
                    elif self.act == "libido":
                        lib += dice(6)
                    elif self.act == "sensitivity":
                        sen += dice(6)
                
                elif self.result == "moderate": # Breaking handled here (no sex happens)
                    
                    if girl.magic_training == "positive":
                        l += 0.5
                    elif girl.magic_training == "negative":
                        f += 0.5
                    
                    if self.act == "obedience":
                        ob += dice(3)
                    elif self.act == "libido":
                        lib += dice(3)
                    elif self.act == "sensitivity":
                        sen += dice(3)
                    else:
                        if girl.magic_training == "positive":
                            brk[self.act] += girl.raise_preference(self.act, "love", 1)
                        elif girl.magic_training == "negative":
                            brk[self.act] += girl.raise_preference(self.act, "fear", 1)
                        else:
                            brk[self.act] += girl.raise_preference(self.act, None, 1)
                        
                # Magic failure
                
                else:
                    if girl.is_("very dom"):
                        l -= 1
                    elif girl.is_("dom"):
                        l -= 0.5
                    elif girl.is_("very sub"):
                        f += 1
                    elif girl.is_("sub"):
                        f += 0.5
                    
                    girl.personality_unlock["DS"] += 15 + MC.get_charisma() + dice(10)
            
            elif self.topic.type == "react" and self.type in ("praise", "gold", "gift", "pet", "day off", "sex"):
                
                ## 1. Personality unlocking
                    
                if self.type == "pet" or self.type == "day off":
                    unlock = "DS"
                elif self.type == "sex":
                    unlock = "LR"
                elif self.type == "praise" or self.type == "gold" or self.type == "gift":
                    unlock = "MI"
                else:
                    raise AssertionError, "Type not found. Type is " + str(self.type)
                
                girl.personality_unlock[unlock] += 15 + MC.get_charisma() + dice(10) # Temp, see how it behaves
                    

                ## 2. Mood, love and fear changes: A score is calculated according to the impact and result of the interaction
                
                m += self.score
                l += self.score
                f -= self.score
                
                if self.score >= 1: # Fear diminishes faster for dom girls
                    if girl.is_("very dom"):
                        f -= 1
                    elif girl.is_("dom"):
                        f -= 0.5
                    elif girl.is_("very sub"):
                        f += 1
                    elif girl.is_("sub"):
                        f += 1
                        
                if self.type == "praise": # More efficient for idealist girls
                    if girl.is_("very idealist"):
                        l += 1
                    elif girl.is_("idealist"):
                        l += 0.5
                    elif girl.is_("very materialist"):
                        l -= 1
                    elif girl.is_("materialist"):
                        l -= 0.5
                
                elif self.type == "gold": # More efficient for materialist girls
                    if girl.is_("very materialist"):
                        l += 1
                    elif girl.is_("materialist"):
                        l += 0.5
                    elif girl.is_("very idealist"):
                        l -= 1
                    elif girl.is_("idealist"):
                        l -= 0.5
                        
                elif self.type == "gift": # Good for all girls
                        l += 1
                
                elif self.type == "pet": # More efficient for sub girls
                    if girl.is_("very sub"):
                        l += 1
                    elif girl.is_("sub"):
                        l += 0.5
                    elif girl.is_("very dom"):
                        l -= 1
                    elif girl.is_("dom"):
                        l -= 0.5
                
                elif self.type == "day off": # More efficient for dom girls
                    if girl.is_("very dom"):
                        l += 1
                    elif girl.is_("dom"):
                        l += 0.5
                    elif girl.is_("very sub"):
                        l -= 1
                    elif girl.is_("sub"):
                        l -= 0.5

                else: # Sexual reward
                    if self.response == "accepted":
                        if girl.is_("very lewd"):
                            l += 2
                        elif girl.is_("lewd"):
                            l += 1
                        elif girl.is_("repressed"):
                            f -= 1
                        
                    elif self.response == "resisted":
                        ne += 1
                        
                        if girl.is_("very dom"):
                            l -= 2
                        elif girl.is_("dom"):
                            l -= 1
                        elif girl.is_("very sub"):
                            l += 1
                        elif girl.is_("sub"):
                            l += 0.5
                        
                        if girl.is_("very repressed"):
                            l -= 1
                            f += 1
                        elif girl.is_("repressed"):
                            l -= 0.5
                            f += 0.5
                        
                    else: # (forced) --> Score is then set to zero
                        ev += 1
                        
                        if girl.is_("very dom"):
                            l -= 2
                        elif girl.is_("dom"):
                            l -= 1
                        elif girl.is_("very sub"):
                            l += 0.5
                            
                        if girl.is_("very repressed"):
                            f += 3
                        elif girl.is_("repressed"):
                            f += 2
                        else:
                            f += 1
                
                # Justify effect
                
                if self.reason: # Idealist girls like fairness
                    if girl.is_("very idealist"):
                        l += 1
                    elif girl.is_("idealist"):
                        l += 0.5
                    elif girl.is_("very materialist"):
                        l -= 0.5
                        
                    ob += self.score
                        
                else: # Materialist girls like to be rewarded for no reason
                    if girl.is_("very materialist"):
                        l += 1
                    elif girl.is_("materialist"):
                        l += 0.5
                    elif girl.is_("very idealist"):
                        l -= 0.5
                
                ## 3. Remember this
                
                girl.will_remember("reward", self.reason, self.score)
                
                
                ## 4. Spoil her
                
                if self.reason:
                    girl.spoil(self.score)
                else:
                    girl.spoil(3*self.score)
        
                
                
            elif self.topic.type == "react" and self.type in ("scold", "deprive", "naked", "beat", "rape", "farm"):
                
                # This happens if a sub girl begs for mercy and MC complies
                if self.MC_reaction == "give up":
                    l += 1
                    f -= 2
                    ev -= 1
                    ob -= dice(6)
                elif self.MC_reaction == "warning":
                    f += 1
                    ne += 1
                    ob -= dice(3)
                else:
                    ## 1. Personality unlocking
                    
                    if self.MC_reaction == "proceed": # Punishment is especially efficient is girl begged and was denied
                        self.score *= 1.5
                        ev += 1
                    
                    if self.type == "beat" or self.type == "farm":
                        unlock = "DS"
                    elif self.type == "naked" or self.type == "rape":
                        unlock = "LR"
                    elif self.type == "scold" or self.type == "deprive":
                        unlock = "MI"
                    else:
                        raise AssertionError, "Type not found. Type is " + str(self.type)
                    
                    girl.personality_unlock[unlock] += 15 + MC.get_charisma() + dice(10) # Temp, see how it behaves
                    
                    ## 2. Mood, love and fear changes
                                        
                    m -= self.score
                    l -= self.score
                    f += self.score
                    ob += self.score
                    
                    if self.score > 0: # Fear increases faster for sub girls
                        if girl.is_("very dom"):
                            f -= 1
                        elif girl.is_("dom"):
                            f -= 0.5
                        elif girl.is_("very sub"):
                            f += 1
                        elif girl.is_("sub"):
                            f += 1
                            
                    if self.type == "scold": # Idealist girls care more about it
                        if girl.is_("very idealist"):
                            f += 1
                        elif girl.is_("idealist"):
                            f += 0.5
                        elif girl.is_("very materialist"):
                            f -= 1
                        elif girl.is_("materialist"):
                            f -= 0.5
                    
                    elif self.type == "deprive": # Hurts materialist girls more
                        if girl.is_("very materialist"):
                            f += 1
                        elif girl.is_("materialist"):
                            f += 0.5
                        elif girl.is_("very idealist"):
                            f -= 1
                        elif girl.is_("idealist"):
                            f -= 0.5
                    
                    elif self.type == "beat": # Scares sub girls more
                        gd -= 1
                        
                        if girl.is_("very sub"):
                            f += 1
                        elif girl.is_("sub"):
                            f += 0.5
                        elif girl.is_("very dom"):
                            f -= 1
                        elif girl.is_("dom"):
                            f -= 0.5
                    
                    elif self.type == "naked": # Repressed girls care more
                        if girl.is_("very repressed"):
                            f += 1
                        elif girl.is_("repressed"):
                            f += 0.5
                        elif girl.is_("very lewd"):
                            f -= 1
                        elif girl.is_("lewd"):
                            f -= 0.5

                    elif self.type == "rape": # Sexual punishment (forced)
                        
                        ev += 1
                        
                        if girl.is_("very dom"):
                            l -= 2
                        elif girl.is_("dom"):
                            l -= 1
                        elif girl.is_("very sub"):
                            l += 0.5
                            
                        if girl.is_("very repressed"):
                            f += 3
                        elif girl.is_("repressed"):
                            f += 2
                        else:
                            f += 1
                            
                        if self.result == "neg_fix":
                            l -= 2
                            f += 3
                        elif self.result == "pos_fix":
                            l += 1
                            f -= 1
                            
                    elif self.type == "farm": # More effective on dom girls, sub girls are somewhat less affected
                        gd -= 1
                        
                        if girl.is_("very dom"):
                            f += 1
                        elif girl.is_("dom"):
                            f += 0.5
                        elif girl.is_("very sub"):
                            f -= 1
                        elif girl.is_("sub"):
                            f -= 0.5
                            
                            
                    ## 3. Justify effect
                    
                    if self.reason: # Idealist girls like fairness
                        if girl.is_("very idealist"):
                            f -= 1
                            ob += 1
                        elif girl.is_("idealist"):
                            f -= 0.5
                            ob += 0.5
                        elif girl.is_("very materialist"):
                            f += 0.5
                            ob -= 0.5
                            
                    else: # Idealist girls hate to be punished for no reason
                        if girl.is_("very idealist"):
                            l -= 1
                            ob -= 1
                        elif girl.is_("idealist"):
                            l -= 0.5
                            ob -= 0.5
                        elif girl.is_("very materialist"):
                            l += 0.5
                            ob += 0.5
                
                
                ## 4. Remembering and terrify effect (applies regardless of whether the punishment was given or not)
                
                # Remember this
                
                girl.will_remember("punishment", self.reason, self.score)
                
                # Terrify her
                
                if self.reason:
                    girl.terrify(self.score)
                else:
                    girl.terrify(3*self.score)
        
        
            ## SEX STAT CHANGES
            
            # Random skill increases
            
            if self.act in extended_sex_acts:
                
                if self.mode == "lecture": # Results of lectures are handled here (not in sexual training)
                                        
                    # Mood and love change
                    if (girl.is_("very lewd") and compare_preference(girl, self.act, "a little reluctant")) or (girl.is_("lewd") and compare_preference(girl, self.act, "a little interested")):
                        l += 2
                        m += 2
                    elif (girl.is_("very lewd") and compare_preference(girl, self.act, "very reluctant")) or (girl.is_("lewd") and compare_preference(girl, self.act, "reluctant")):
                        m += 1
                    elif (girl.is_("very repressed") and compare_preference(girl, self.act, "a little interested")) or (girl.is_("repressed") and compare_preference(girl, self.act, "a little reluctant")):
                        m -= 1
                    elif (girl.is_("very repressed") and compare_preference(girl, self.act, "reluctant")) or (girl.is_("repressed") and compare_preference(girl, self.act, "very reluctant")):
                        l -= 2
                        m -= 2
                    
                    if (girl.get_love() - girl.get_fear()) // 10 >= 0:
                        mod = "love"
                    else:
                        mod = "fear"
                    
                    if self.score > 0:
                        brk[self.act] += girl.raise_preference(self.act, mod, 2)
                        
                    elif self.score == 0:
                        brk[self.act] += girl.raise_preference(self.act, mod, 1)
                    
                    else:
                        brk[self.act] += girl.raise_preference(self.act, mod, 0.25)
                
                elif self.topic.label == "slave_remove_fixation":
                    if self.result != "locked": # Modest break even if training fails
                        brk[self.act] + girl.raise_preference(self.act, self.MC_reaction, 0.5)
                    else: # Negative break if fix is locked
                        brk[self.act] + girl.raise_preference(self.act, self.MC_reaction, -1)
                    
                    # Prestige
                    
                    if self.act == "naked":
                        p += girl.rank / 2
                    else:
                        p += girl.rank
                    
                    
                else:
                
                    if not (self.MC_reaction in ("give up", "warning") or self.result in ("fail", "fled", "moderate")):
                        
                        if self.act == "naked":

                            lib += dice(6)
                            
                            d = dice(6)
                            
                            if d >= 6:
                                ob += dice(3)
                            elif d >= 5:
                                bea += dice(3)
                            elif d >= 4:
                                bod += dice(3)
                                
                        elif self.act == "service":
                            
                            sv += dice(6)
                            
                            d = dice(6)
                            
                            if d >= 5:
                                sen += dice(3)
                            elif d >= 4:
                                cha += dice(3)
                        
                        elif self.act == "sex":
                            
                            sx += dice(6)
                            
                            d = dice(6)
                            
                            if d >= 5:
                                lib += dice(3)
                            elif d >= 4:
                                bea += dice(3)
                        
                        elif self.act == "anal":
                            
                            an += dice(6)
                            
                            d = dice(6)
                            
                            if d >= 5:
                                con += dice(3)
                            elif d >= 4:
                                bod += dice(3)
                                
                        elif self.act == "fetish":
                            
                            fe += dice(6)
                            
                            d = dice(6)
                            
                            if d >= 5:
                                ob += dice(3)
                            elif d >= 4:
                                ref += dice(3)
                                
                        elif self.act == "bisexual":
                            
                            if dice(6) >= 4:
                                sx += dice(6)
                            else:
                                sv += dice(6)
                            
                            d = dice(6)
                            
                            if d >= 5:
                                sen += dice(3)
                            elif d >= 4:
                                lib += dice(3)
                                
                        elif self.act == "group":
                            
                            d = dice(6)
                            
                            if d >= 5:
                                sv += dice(6)
                            elif d >= 3:
                                sx += dice(6)
                            else:
                                an += dice(6)
                            
                            d = dice(6)
                            
                            if d >= 5:
                                con += dice(3)
                            elif d >= 4:
                                ob += dice(3)
                        
                        # Breaking (Warning: lecture breaking is handled above)
                        
                        if self.topic.type == "magic":
                            if girl.magic_training == "positive":
                                brk[self.act] += girl.raise_preference(self.act, "love", 3)
                            elif girl.magic_training == "negative":
                                brk[self.act] += girl.raise_preference(self.act, "fear", 3)
                            else:
                                brk[self.act] += girl.raise_preference(self.act, None, 3)
                        
                        elif self.response == "accepted":
                            brk[self.act] += girl.raise_preference(self.act, "love", 3)
                        
                        elif self.response == "resisted":
                            if girl.is_("sub"):
                                brk[self.act] += girl.raise_preference(self.act, "fear", 2)
                            elif girl.is_("dom"):
                                brk[self.act] += girl.raise_preference(self.act, "fear", 1)
                        
                        elif self.response == "refused":
                            if girl.is_("sub"):
                                brk[self.act] += girl.raise_preference(self.act, "fear", 3)
                            elif girl.is_("dom"):
                                brk[self.act] += girl.raise_preference(self.act, "fear", -1)
                            
                        else:
                            brk[self.act] += girl.raise_preference(self.act, None, 3)
                            
                        # Prestige
                        
                        if self.act != "naked":
                            brk["naked"] += girl.raise_preference("naked", None, 1) # Naked acts rise with other sex acts
                            p += girl.rank
                        
                        if self.act in ("sex", "group"):
                            if girl.pop_virginity(origin="MC"):
                                virgin = True
                                p += 3 * girl.rank
                        
                        
            ## Apply changes
            
            changes = [("mood", m), ("love", l), ("fear", f), ("beauty", bea), ("body", bod), ("charm", cha), ("refinement", ref), ("obedience", ob), ("sensitivity", sen), ("constitution", con), ("libido", lib), ("service", sv), ("sex", sx), ("anal", an), ("fetish", fe)]
            
#            renpy.say("", "love value:" + str(l))
            
            text1 = ""
            
            for s, v in changes: # Charisma impacts all changes
                
                c = girl.change_stat(s, v * (1 + (MC.get_charisma()/10)))
                
                if debug_mode:
                    shown = str(round_int(c))
                else:
                    shown = get_plus_rating(c)
                
                if v != 0 and c != 0:
                    text1 += "\n" + s.capitalize() + ": " + shown
                                        
            text1 += "\n"
            
            for a in brk.keys():
                                
                if brk[a] != None:
                    if debug_mode:
                        shown = str(round_int(brk[a]))
                    else:
                        shown = get_plus_rating(brk[a], "pref")
                                        
                    text1 += "\n" + a.capitalize() + " preference: " + shown
                else:
                    raise AssertionError, "Unexpected breaking value for " + a + ". Please report this bug."
                
            if virgin:
                text1 += "\n" + girl.fullname + " has lost her virginity with you."
            
            text1 += "\n"
            
            if gd:
                MC.good += gd
            if ne:
                MC.neutral += ne
            if ev:
                MC.evil += ev
            
            if debug_mode:
                if gd:
                    text1 += "\ngood: " + str(gd)
                if ne:
                    text1 += "\nneutral: " + str(ne)
                if ev:
                    text1 += "\nevil: " + str(ev)
            if p:
                MC.change_prestige(p)
                text1 += "\nprestige: " + str(p)

            if not text2:
                text1 = "No changes"
            renpy.call_screen("OK_screen", girl.fullname + " - Interaction results", text1, dark=True)
            
            return
        
            
            
    class GirlRecentEvent(object):
        
        def __init__(self, type, action=None, base_description="", encourage=True, discipline=True):
            
            self.type = type
            self.action = action
            self.base_description = base_description
            self.description = ""
            self.encourage = encourage
            self.discipline = discipline
            
            self.time = 0
            self.rewarded = 0
            self.punished = 0
            
        def reward(self, score):
            
            self.rewarded += score
            
            if self.rewarded > 10:
                self.rewarded = 10
            
        def punish(self, score):
            
            self.punished += score
            
            if self.punished > 10:
                self.punished = 10
            
        def refresh(self):
            if self.rewarded > 0:
                self.rewarded -= 1
            if self.punished > 0:
                self.punished -= 1
            
        
    class Mod(object):
    
        """This class is used to track external mods. Mods are declared in their respective rpy files, and automatically added to 'detected_mods' upon creation."""
        
        def __init__(self, name, folder, creator="Unknown", version= 1.0, pic=None, description="This is a mod for Brothel King.", help_prompts=[], init_label="", night_label = "", events={}):        
                          
            #### Init variables
            self.name = name
            self.path = "mods/" + folder + "/"
            self.creator = creator
            self.version = version
            if pic:
                self.pic = Picture(pic, self.path + pic)
            else:
                self.pic = None
            self.description = description
            self.full_name = name + " v" + str(self.version) + ", from " + creator
            
            ## help_prompts is a list of tuples (name, label), each representing a menu prompt in the 'help/mod/mod options' menu. 
            ## 'name' is the prompt message as it appears on the menu button, and 'label' is the target label it will call (not jump).
            ## Several actions can be added to the list by adding tuples (name, label) to the list, each separated by a comma
            self.help_prompts = []
            ## Adds [mod name] to prompt messages for clarity
            for prompt in help_prompts:
                self.help_prompts.append(("[[" + self.name + "] " + prompt[0], prompt[1]))  
            
            ## Init label: This will run when the mod is started, allowing to set some variables and events if necessary
            self.init_label = init_label
            self.night_label = night_label
            
            ## Event dictionary (all mod events must be declared here)
            self.events = events
            
            
            #### Default variables
            self.flags = defaultdict(bool)
            
            for ev in self.events.values():
                ev.mod = self
            
            detected_mods[self.name] = self
            self.seen = False
            self.active = False
        
        def get_check(self): # Returns a list of values used for comparison 
            return (self.version, self.path, self.init_label, len(self.events))
        
        def check_for_updates(self): # Checks a list of values used for comparison with existing mods
            
            global mod_traceback
            
            if self.get_check() != persistent.mods[self.name]["check"]:
#                mod_traceback += "Checked " + str(self.get_check()) + " against " + str(persistent.mods[self.name]["check"])
                return True
            return False
                        
        def activate(self):
            if renpy.call_screen("yes_no", "Do you want to activate " + self.full_name + "?"):
                if not self.active or not persistent.mods[self.name]["active"]:
                    self.active = True
                    persistent.mods[self.name]["active"] = True
                    renpy.notify(self.name + " has been activated.")
                    reset_updated_games()
                    
                else:
                    renpy.notify(self.name + " is already active.")
                
        def deactivate(self):
            if renpy.call_screen("yes_no", "Do you really want to deactivate " + self.full_name + "? This might negatively affect games saved while this mod was on."):
                if self.active or persistent.mods[self.name]["active"]:
                    self.active = False
                    persistent.mods[self.name]["active"] = False
                    renpy.notify(self.name + " has been deactivated.")
                    reset_updated_games()
                else:
                    renpy.notify(self.name + " couldn't be found among active mods.")
        
        def add_event(self, event_name, type=None, time=None, delay=1, call_args=[]): # event_name is the event label (not object). time is the exact calendar date. If not provided, current time + delay is used instead (D+1 by default).
            
            if call_args:
                self.events[event_name].call_args = call_args
            
            if type:
                self.events[event_name].type = type
            else:
                type = self.events[event_name].type 
            
            if type == "alarm":
                if not time:
                    time = calendar.time + delay
                calendar.set_alarm(time, self.events[event_name])
                
                if time <= calendar.time:
                    renpy.say("System", "Warning: Event set to a past date. Change the event time or delay.")
            
            elif type in ("morning", "day", "night"):
                daily_events.append(self.events[event_name])
            
            elif type == "city":
                city_events.append(self.events[event_name])
        
        def set_condition(self, condition, value):
            self.flags[condition] = value
    
    class MC_challenge(object):
        """This class is used to run Player challenges and return a result."""
        
        def __init__(self, name, stat, opposed):
            
            self.name = name
            self.pic = Picture(name + ".jpg", "UI/challenges/" + name + ".jpg")
            self.stat = stat
            self.opposed = opposed
            self.d = 0 # stores the latest MC dice roll
            self.d_op = 0 # stores the latest opponent dice roll
            
            self.score = 0 # stores the latest MC score
            self.score_op = 0 # stores the latest opponent score
            
            self.result = 0 # stores the result of the latest challenge
        
        def get_pic(self, x, y):
            return self.pic.get(x, y)
        
        def get_score(self, diff, bonus=0, opponent_bonus=0, dice_faces=6, dice_nb = 1, raw=False):
            
            self.score = MC.get_stat(self.stat, raw=raw) + MC.get_effect("change", self.name + " challenges")
            
            self.d = dice(dice_faces, dice_nb)
            self.score += self.d + bonus
            
            if self.opposed:
                self.d_op = dice(dice_faces, dice_nb)
                self.score_op = diff + self.d_op + opponent_bonus
                
                result = self.score - self.score_op
            
            else:
                result = self.score - diff
            
            return result
            
        def run(self, diff, score=False, raw=False, bonus=0, opponent_bonus=0, dice_faces=6, dice_nb = 1, strict=False):
            
            if not score:
                return self.run_pass(diff, bonus, opponent_bonus, dice_faces, dice_nb, raw, strict)
            else:
                return self.get_score(diff, bonus, opponent_bonus, dice_faces, dice_nb, raw)
        
        def run_pass(self, diff, bonus=0, opponent_bonus=0, dice_faces=6, dice_nb=1, raw=False, strict=False): # This test returns pass or fail. strict tests require a strictly superior result to succeed.
            
            self.result = self.get_score(diff, bonus, opponent_bonus, dice_faces, dice_nb, raw)
            
            if self.result > 0:
                return True
            elif self.result < 0:
                return False
            elif strict:
                return False
            else:
                return True
                
        def estimate_diff(self, diff, raw=False, score=False, bonus=0, opponent_bonus=0, strict=False, percentage=False): # Returns difficulty as a qualifier or percentage
            
            differential = (diff + opponent_bonus) - (MC.get_stat(self.stat, raw) + bonus + MC.get_effect("change", self.name + " challenges"))
            
            if not strict:
                differential -= 1
            
            if not self.opposed:
                chance = 1.0 - (differential)/6.0
            else:
                for d, c in [(5, 0.0), (4, 0.03), (3, 0.08), (2, 0.17), (1, 0.28), (0, 0.42), (-1, 0.58), (-2, 0.72), (-3, 0.83), (-4, 0.92), (-99, 1.0)]:
                    if differential >= d:
                        chance = c
                        break
                else:
                    raise AssertionError("Couldn't process differential during MC challenge.")
            
            if percentage:
                return str(int(chance*100)) + "%"
            
            else:
                if chance >= 1.0 and not score:
                    return "Safe"
                elif chance > 0.8:
                    return "Very easy"
                elif chance > 0.6:
                    return "Easy"
                elif chance > 0.4:
                    return "Fair"
                elif chance > 0.2:
                    return "Hard"
                elif chance > 0.0:
                    return "Very hard"
                else:
                    return "Impossible"
            
    class Resource(object):
        
        """Resources are sold at the market and extracted in the city. They serve primarily for furniture and city events."""
        
        def __init__(self, name, rank, stat=None, sound=s_gold, description="", location=None):
            self.name = name
            self.rank = rank
            self.stat = stat
            self.pic = Picture(name + ".webp", "UI/resources/" + name + ".webp")
            self.sound = sound
            self.description = description
            self.location = location
            
        def get_pic(self, x, y):
            return self.pic.get(x, y)
            
        def activate_extractor(self):
            auto_extractors[self.name] = True
            auto_extractors[self.name + " ON"] = True
            auto_extractors[self.name + " durability"] = 150
            self.location.menu = (self.location.menu[0] + " [[E]", self.location.menu[1])
        
    class Furniture(object):
        
        """Furniture are upgrades for the brothel that provide permanent bonuses."""
        
        def __init__(self, name, type, pic=None, rank=2, chapter=2, cost=[], duration=0, effects=[], base_description="", upgrade=False):
            self.name = name
            self.type = type
            if pic:
                self.pic = Picture(pic, "items/furniture/" + pic)
            else:
                self.pic = Picture("misc.webp", "items/misc/misc.webp")
            self.rank = rank # This is the rank at which the furniture becomes available
            self.chapter = chapter # This is the chapter at which the furniture becomes available (if not rank)
            self.cost = cost # Cost is a list of tuples (resource=str, amount=int)
            self.duration = duration
            self.effects = effects
            self.description = "{b}" + self.name + "{/b}" + ": " + get_description(base_description, effects)
            self.upgrade = upgrade
            
            if rank == 0: # 0 rank furniture is built from the beginning
                self.built = True
                self.active = True
                brothel.furniture.append(self)
            else:
                self.built = False
                self.active = False
            
        def get_pic(self, x, y):
            return self.pic.get(x, y)
        
        def can_build(self):
            if brothel.current_building == self:
                return False
            elif self.built or (self.rank > district.rank) or (self.chapter > game.chapter):
                return False
            elif self.upgrade:
                if not furniture_dict[self.upgrade].built:
                    return False
            return True
        
        def start_building(self):
            if self.duration:
                renpy.say(carpenter, "I'll be finished in " + str(self.duration) + " days. I'm sure you'll be happy with the result.")
                if not story_flags["carpenter first build"]:
                    calendar.set_alarm(calendar.time+1, StoryEvent(label="iulia1", type="morning"))
                    story_flags["carpenter first build"] = True
                calendar.set_alarm(calendar.time + self.duration, Event(label = "furniture_built", object=self))
                brothel.current_building = self
                brothel.started_building = calendar.time
            else:
                self.build()
                
        def build(self, message=True):
            self.built = True
            brothel.furniture.append(self)
            if self.upgrade:
                brothel.furniture.remove(furniture_dict[self.upgrade])
                brothel.deactivate_furniture(furniture_dict[self.upgrade])
                if message:
                    renpy.call_screen("OK_screen", title = "Furniture Upgraded", message = self.upgrade + " has been upgraded to a " + self.name + ".\n\n" + self.description, pic = self.pic, pic_size = "small")
            elif message:
                renpy.call_screen("OK_screen", title = "Furniture Built", message = "A new " + self.name + " has been built.\n\n" + self.description, pic = self.pic, pic_size = "small")
            self.activate()
            
        def destroy(self, message=True):
            self.built = False
            if self in brothel.furniture:
                brothel.furniture.remove(self)
            self.deactivate()
            if self.upgrade:
                brothel.furniture.append(furniture_dict[self.upgrade])
                brothel.activate_furniture(furniture_dict[self.upgrade])
                if message:
                    renpy.call_screen("OK_screen", title = "Furniture destroyed", message = self.name + " has been destroyed and replaced with " + self.upgrade, pic = self.pic, pic_size = "small")
            elif message:
                renpy.call_screen("OK_screen", title = "Furniture destroyed", message = self.name + " has been destroyed.", pic = self.pic, pic_size = "small")
            self.activate()
        
        def activate(self):
            if not self.active and self.built:
                self.active = True
                update_effects()
                return True
            return False
                
        def deactivate(self):
            if self.active:
                self.active = False
                update_effects()
                return True
            return False
        
        def toggle(self):
            if self.active:
                self.deactivate()
            else:
                self.activate()
                
        def describe_cost(self):
            dlist = [(str(amount) + " " + resource) for resource, amount in self.cost]
            
            return and_text(dlist)
                        
            
    class Moon(object):
        
        """A new Moon appears every month. They have an effect on gameplay."""
        
        def __init__(self, name, effects=[], description="", sound=None):
            self.name = name.capitalize() + " Moon"
            self.pic = Picture(name + ".webp", "backgrounds/moons/" + name + ".webp")
            self.effects = effects
            self.short_description = get_description(self.name, self.effects, separator=": ")
            self.description = get_description(description, self.effects)
            self.sound = sound
            
        def get_pic(self, x, y):
            return self.pic.get(x, y)
    
    
    class Loan(object):
        
        """A loan is a sum that must be paid back to the banker, with or without interest"""
        
        def __init__(self, amount, interest=0, duration=10): # Where interest is a float (interest_rate)
            self.amount = amount
            self.initial_amount = amount
            self.interest = interest
            self.duration = duration
            self.total_cost = int(amount * (1+interest))
            self.daily_cost = int(self.total_cost / duration)
            
        def repay(self):
            amount_reimbursed = get_change_min_max(self.amount, -self.initial_amount/self.duration, 0)
            interest_reimbursed = amount_reimbursed * self.interest
                        
            MC.gold += int(amount_reimbursed + interest_reimbursed)
            self.amount += amount_reimbursed
            
            return -int(amount_reimbursed + interest_reimbursed)
            
        
            
## UTILITY CLASSES  ##
## Not mine !       ##


    class ProportionalScale(im.ImageBase):
        '''Resizes a renpy image to fit into the specified width and height.
        The aspect ratio of the image will be conserved.'''
        def __init__(self, imgname, maxwidth, maxheight, bilinear=True, **properties):
            img = im.image(imgname)
            super(ProportionalScale, self).__init__(img, maxwidth, maxheight, bilinear, **properties)
            self.imgname = imgname
            self.image = img
            self.maxwidth = int(maxwidth)
            self.maxheight = int(maxheight)
            self.bilinear = bilinear

        def load(self):
            #<Chris12 NotFound>
            # Loads a neutral image instead of failing, in case an image does not exist
            try :
                child = im.cache.get(self.image)
            except IOError :
                renpy.notify("Missing: " + self.imgname)
                child = im.cache.get(Image("backgrounds/not_found.webp"))
            #</Chris12 NotFound>

            width, height = child.get_size()
            
            ratio = min(self.maxwidth/float(width), self.maxheight/float(height))
            width = ratio * width
            height = ratio * height

            if self.bilinear:
                try:
                    renpy.display.render.blit_lock.acquire()
                    rv = renpy.display.scale.smoothscale(child, (width, height))
                finally:
                    renpy.display.render.blit_lock.release()
            else:
                try:
                    renpy.display.render.blit_lock.acquire()
                    rv = renpy.display.pgrender.transform_scale(child, (newwidth, newheight))
                finally:
                    renpy.display.render.blit_lock.release()
            return rv

        def predict_files(self):
            return self.image.predict_files()

            
            